import unittest
import requests_mock
from parsons import TargetSmartAPI, Table
from test.utils import validate_list
from test.responses.ts_responses import (
    address_response,
    district_point,
    district_expected,
    district_zip,
    zip_expected,
    phone_response,
    phone_expected,
    radius_response,
)

output_list = [
    {
        "vb.tsmart_zip": "60625",
        "vb.vf_g2014": "Y",
        "vb.vf_g2016": "Y",
        "vb.tsmart_middle_name": "H",
        "ts.tsmart_midterm_general_turnout_score": "85.5",
        "vb.tsmart_name_suffix": "",
        "vb.voterbase_gender": "Male",
        "vb.tsmart_city": "CHICAGO",
        "vb.tsmart_full_address": "908 N MAIN AVE APT 2",
        "vb.voterbase_phone": "5125705356",
        "vb.tsmart_partisan_score": "99.6",
        "vb.tsmart_last_name": "BLANKS",
        "vb.voterbase_id": "IL-12568670",
        "vb.tsmart_first_name": "BILLY",
        "vb.voterid": "Q8W8R82Z",
        "vb.voterbase_age": "37",
        "vb.tsmart_state": "IL",
        "vb.voterbase_registration_status": "Registered",
    }
]


class TestTargetSmartAPI(unittest.TestCase):
    def setUp(self):

        self.ts = TargetSmartAPI(api_key="FAKEKEY")

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_data_enhance(self, m):

        json = {
            "input": {"search_id": "IL-12568670", "search_id_type": "voterbase"},
            "error": None,
            "output": output_list,
            "output_size": 1,
            "match_found": True,
            "gateway_id": "b8c86f27-fb32-11e8-9cc1-45bc340a4d22",
            "function_id": "b8c98093-fb32-11e8-8b25-e99c70f6fe74",
        }

        expected = [
            "vb.tsmart_zip",
            "vb.vf_g2014",
            "vb.vf_g2016",
            "vb.tsmart_middle_name",
            "ts.tsmart_midterm_general_turnout_score",
            "vb.tsmart_name_suffix",
            "vb.voterbase_gender",
            "vb.tsmart_city",
            "vb.tsmart_full_address",
            "vb.voterbase_phone",
            "vb.tsmart_partisan_score",
            "vb.tsmart_last_name",
            "vb.voterbase_id",
            "vb.tsmart_first_name",
            "vb.voterid",
            "vb.voterbase_age",
            "vb.tsmart_state",
            "vb.voterbase_registration_status",
        ]

        m.get(self.ts.connection.uri + "person/data-enhance", json=json)

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.ts.data_enhance("IL-12568678")))

        # Assert exception on missing state
        with self.assertRaises(Exception):
            self.ts.data_enhance("vb0001", search_id_type="votebuilder")

        # Assert exception on missing state
        with self.assertRaises(Exception):
            self.ts.data_enhance("vb0001", search_id_type="smartvan")

        # Assert exception on missing state
        with self.assertRaises(Exception):
            self.ts.data_enhance("vb0001", search_id_type="voter")

        # Assert works with state provided
        for i in ["votebuilder", "voter", "smartvan"]:
            self.assertTrue(
                validate_list(
                    expected,
                    self.ts.data_enhance("IL-12568678", search_id_type=i, state="IL"),
                )
            )

    @requests_mock.Mocker()
    def test_radius_search(self, m):

        m.get(self.ts.connection.uri + "person/radius-search", json=radius_response)

        expected = [
            "similarity_score",
            "distance_km",
            "distance_meters",
            "distance_miles",
            "distance_feet",
            "proximity_score",
            "composite_score",
            "uniqueness_score",
            "confidence_indicator",
            "ts.tsmart_midterm_general_turnout_score",
            "vb.tsmart_city",
            "vb.tsmart_first_name",
            "vb.tsmart_full_address",
            "vb.tsmart_last_name",
            "vb.tsmart_middle_name",
            "vb.tsmart_name_suffix",
            "vb.tsmart_partisan_score",
            "vb.tsmart_precinct_id",
            "vb.tsmart_precinct_name",
            "vb.tsmart_state",
            "vb.tsmart_zip",
            "vb.tsmart_zip4",
            "vb.vf_earliest_registration_date",
            "vb.vf_g2014",
            "vb.vf_g2016",
            "vb.vf_precinct_id",
            "vb.vf_precinct_name",
            "vb.vf_reg_cass_address_full",
            "vb.vf_reg_cass_city",
            "vb.vf_reg_cass_state",
            "vb.vf_reg_cass_zip",
            "vb.vf_reg_cass_zip4",
            "vb.vf_registration_date",
            "vb.voterbase_age",
            "vb.voterbase_gender",
            "vb.voterbase_id",
            "vb.voterbase_phone",
            "vb.voterbase_registration_status",
            "vb.voterid",
        ]

        # Assert response is expected structure
        def rad_search():
            return self.ts.radius_search(
                "BILLY",
                "Burchard",
                radius_size=100,
                address="908 N Washtenaw, Chicago, IL",
            )

        self.assertTrue(validate_list(expected, rad_search()))

    def test_district_args(self):

        self.assertRaises(ValueError, self.ts.district, search_type="address")
        self.assertRaises(ValueError, self.ts.district, search_type="zip", zip4=9)
        self.assertRaises(ValueError, self.ts.district, search_type="zip", zip5=0)
        self.assertRaises(ValueError, self.ts.district, search_type="point")
        self.assertRaises(ValueError, self.ts.district, search_type="zip")

    @requests_mock.Mocker()
    def test_district_point(self, m):

        # Test Points
        m.get(self.ts.connection.uri + "service/district", json=district_point)
        self.assertTrue(
            validate_list(
                district_expected,
                self.ts.district(
                    search_type="point", latitude="41.898369", longitude="-87.694382"
                ),
            )
        )

    @requests_mock.Mocker()
    def test_district_zip(self, m):
        # Test Zips
        m.get(self.ts.connection.uri + "service/district", json=district_zip)
        self.assertTrue(
            validate_list(
                zip_expected,
                self.ts.district(search_type="zip", zip5="60622", zip4="7194"),
            )
        )

    @requests_mock.Mocker()
    def test_district_address(self, m):
        # Test Address
        m.get(self.ts.connection.uri + "service/district", json=address_response)
        self.assertTrue(
            validate_list(
                district_expected,
                self.ts.district(
                    search_type="address", address="908 N Main St, Chicago, IL 60611"
                ),
            )
        )

    @requests_mock.Mocker()
    def test_phone(self, m):

        # Test phone
        m.get(self.ts.connection.uri + "person/phone-search", json=phone_response)
        self.assertTrue(
            validate_list(phone_expected, self.ts.phone(Table([{"phone": 4435705355}])))
        )



import os
import unittest
import requests_mock
from unittest import mock
from test.utils import assert_matching_tables
from parsons import Bloomerang, Table

from test.test_bloomerang.test_data import (
    ENV_PARAMETERS,
    ID,
    TEST_DELETE,
    TEST_CREATE_CONSTITUENT,
    TEST_GET_CONSTITUENT,
    TEST_GET_CONSTITUENTS,
    TEST_CREATE_TRANSACTION,
    TEST_GET_TRANSACTION,
    TEST_GET_TRANSACTIONS,
    TEST_CREATE_INTERACTION,
    TEST_GET_INTERACTION,
    TEST_GET_INTERACTIONS,
)


class TestBloomerang(unittest.TestCase):
    def setUp(self):
        self.bloomerang = Bloomerang(api_key="test_key")

    @mock.patch.dict(os.environ, ENV_PARAMETERS)
    def test_init_env(self):
        bloomerang = Bloomerang()
        self.assertEqual(bloomerang.api_key, "env_api_key")
        self.assertEqual(bloomerang.client_id, "env_client_id")
        self.assertEqual(bloomerang.client_secret, "env_client_secret")

    @requests_mock.Mocker()
    def test_authentication(self, m):
        # API key
        bloomerang = Bloomerang(api_key="my_key")
        self.assertEqual(bloomerang.conn.headers["X-API-KEY"], "my_key")

        # OAuth2
        m.post(url=bloomerang.uri_auth, json={"code": "my_auth_code"})
        m.post(
            url=bloomerang.uri + "oauth/token", json={"access_token": "my_access_token"}
        )
        bloomerang = Bloomerang(client_id="my_id", client_secret="my_secret")
        self.assertEqual(bloomerang.authorization_code, "my_auth_code")
        self.assertEqual(bloomerang.access_token, "my_access_token")
        self.assertEqual(
            bloomerang.conn.headers["Authorization"], "Bearer my_access_token"
        )

    def test_base_endpoint(self):
        url = self.bloomerang._base_endpoint("constituent")
        self.assertEqual(url, "https://api.bloomerang.co/v2/constituent/")

        url = self.bloomerang._base_endpoint("constituent", 1234)
        self.assertEqual(url, "https://api.bloomerang.co/v2/constituent/1234/")

        url = self.bloomerang._base_endpoint("constituent", "1234")
        self.assertEqual(url, "https://api.bloomerang.co/v2/constituent/1234/")

    @requests_mock.Mocker()
    def test_create_constituent(self, m):
        m.post(f"{self.bloomerang.uri}constituent/", json=TEST_CREATE_CONSTITUENT)
        self.assertEqual(self.bloomerang.create_constituent(), TEST_CREATE_CONSTITUENT)

    @requests_mock.Mocker()
    def test_update_constituent(self, m):
        m.put(f"{self.bloomerang.uri}constituent/{ID}/", json=TEST_CREATE_CONSTITUENT)
        self.assertEqual(
            self.bloomerang.update_constituent(ID), TEST_CREATE_CONSTITUENT
        )

    @requests_mock.Mocker()
    def test_get_constituent(self, m):
        m.get(f"{self.bloomerang.uri}constituent/{ID}/", json=TEST_GET_CONSTITUENT)
        self.assertEqual(self.bloomerang.get_constituent(ID), TEST_GET_CONSTITUENT)

    @requests_mock.Mocker()
    def test_delete_constituent(self, m):
        m.delete(f"{self.bloomerang.uri}constituent/{ID}/", json=TEST_DELETE)
        self.assertEqual(self.bloomerang.delete_constituent(ID), TEST_DELETE)

    @requests_mock.Mocker()
    def test_get_constituents(self, m):
        m.get(
            f"{self.bloomerang.uri}constituents/?skip=0&take=50",
            json=TEST_GET_CONSTITUENTS,
        )
        assert_matching_tables(
            self.bloomerang.get_constituents(), Table(TEST_GET_CONSTITUENTS["Results"])
        )

    @requests_mock.Mocker()
    def test_create_transaction(self, m):
        m.post(f"{self.bloomerang.uri}transaction/", json=TEST_CREATE_TRANSACTION)
        self.assertEqual(self.bloomerang.create_transaction(), TEST_CREATE_TRANSACTION)

    @requests_mock.Mocker()
    def test_update_transaction(self, m):
        m.put(f"{self.bloomerang.uri}transaction/{ID}/", json=TEST_CREATE_TRANSACTION)
        self.assertEqual(
            self.bloomerang.update_transaction(ID), TEST_CREATE_TRANSACTION
        )

    @requests_mock.Mocker()
    def test_get_transaction(self, m):
        m.get(f"{self.bloomerang.uri}transaction/{ID}/", json=TEST_GET_TRANSACTION)
        self.assertEqual(self.bloomerang.get_transaction(ID), TEST_GET_TRANSACTION)

    @requests_mock.Mocker()
    def test_delete_transaction(self, m):
        m.delete(f"{self.bloomerang.uri}transaction/{ID}/", json=TEST_DELETE)
        self.assertEqual(self.bloomerang.delete_transaction(ID), TEST_DELETE)

    @requests_mock.Mocker()
    def test_get_transactions(self, m):
        m.get(
            f"{self.bloomerang.uri}transactions/?skip=0&take=50",
            json=TEST_GET_TRANSACTIONS,
        )
        assert_matching_tables(
            self.bloomerang.get_transactions(), Table(TEST_GET_TRANSACTIONS["Results"])
        )

    @requests_mock.Mocker()
    def test_get_transaction_designation(self, m):
        m.get(
            f"{self.bloomerang.uri}transaction/designation/{ID}/",
            json=TEST_GET_TRANSACTION,
        )
        self.assertEqual(
            self.bloomerang.get_transaction_designation(ID), TEST_GET_TRANSACTION
        )

    @requests_mock.Mocker()
    def test_get_transaction_designations(self, m):
        m.get(
            f"{self.bloomerang.uri}transactions/designations/?skip=0&take=50",
            json=TEST_GET_TRANSACTIONS,
        )
        assert_matching_tables(
            self.bloomerang.get_transaction_designations(),
            Table(TEST_GET_TRANSACTIONS["Results"]),
        )

    @requests_mock.Mocker()
    def test_create_interaction(self, m):
        m.post(f"{self.bloomerang.uri}interaction/", json=TEST_CREATE_INTERACTION)
        self.assertEqual(self.bloomerang.create_interaction(), TEST_CREATE_INTERACTION)

    @requests_mock.Mocker()
    def test_update_interaction(self, m):
        m.put(f"{self.bloomerang.uri}interaction/{ID}/", json=TEST_CREATE_INTERACTION)
        self.assertEqual(
            self.bloomerang.update_interaction(ID), TEST_CREATE_INTERACTION
        )

    @requests_mock.Mocker()
    def test_get_interaction(self, m):
        m.get(f"{self.bloomerang.uri}interaction/{ID}/", json=TEST_GET_INTERACTION)
        self.assertEqual(self.bloomerang.get_interaction(ID), TEST_GET_INTERACTION)

    @requests_mock.Mocker()
    def test_delete_interaction(self, m):
        m.delete(f"{self.bloomerang.uri}interaction/{ID}/", json=TEST_DELETE)
        self.assertEqual(self.bloomerang.delete_interaction(ID), TEST_DELETE)

    @requests_mock.Mocker()
    def test_get_interactions(self, m):
        m.get(
            f"{self.bloomerang.uri}interactions/?skip=0&take=50",
            json=TEST_GET_INTERACTIONS,
        )
        assert_matching_tables(
            self.bloomerang.get_interactions(), Table(TEST_GET_INTERACTIONS["Results"])
        )



import typing as t
from parsons.utilities import files
from parsons.utilities import check_env
import json
import os


def setup_google_application_credentials(
    app_creds: t.Union[t.Dict, str, None],
    env_var_name: str = "GOOGLE_APPLICATION_CREDENTIALS",
) -> None:
    # Detect if app_creds is a dict, path string or json string, and if it is a
    # json string, then convert it to a temporary file. Then set the
    # environmental variable.
    credentials = check_env.check(env_var_name, app_creds)
    try:
        if type(credentials) is dict:
            credentials = json.dumps(credentials)
        if json.loads(credentials):
            creds_path = files.string_to_temp_file(credentials, suffix=".json")
    except ValueError:
        creds_path = credentials

    os.environ[env_var_name] = creds_path



import os
import json
import logging

from parsons.etl.table import Table
from parsons.google.utitities import setup_google_application_credentials

import gspread
from google.oauth2.service_account import Credentials

logger = logging.getLogger(__name__)


class GoogleSheets:
    """
    A connector for Google Sheets, handling data import and export.

    `Args:`
        google_keyfile_dict: dict
            A dictionary of Google Drive API credentials, parsed from JSON provided
            by the Google Developer Console. Required if env variable
            ``GOOGLE_DRIVE_CREDENTIALS`` is not populated.
        subject: string
            In order to use account impersonation, pass in the email address of the account to be
            impersonated as a string.
    """

    def __init__(self, google_keyfile_dict=None, subject=None):

        scope = [
            "https://spreadsheets.google.com/feeds",
            "https://www.googleapis.com/auth/drive",
        ]

        setup_google_application_credentials(
            google_keyfile_dict, "GOOGLE_DRIVE_CREDENTIALS"
        )
        google_credential_file = open(os.environ["GOOGLE_DRIVE_CREDENTIALS"])
        credentials_dict = json.load(google_credential_file)

        credentials = Credentials.from_service_account_info(
            credentials_dict, scopes=scope, subject=subject
        )

        self.gspread_client = gspread.authorize(credentials)

    def _get_worksheet(self, spreadsheet_id, worksheet=0):
        # Internal method to retrieve a worksheet object.

        # Check if the worksheet is an integer, if so find the sheet by index
        if isinstance(worksheet, int):
            return self.gspread_client.open_by_key(spreadsheet_id).get_worksheet(
                worksheet
            )

        elif isinstance(worksheet, str):
            idx = self.list_worksheets(spreadsheet_id).index(worksheet)
            try:
                return self.gspread_client.open_by_key(spreadsheet_id).get_worksheet(
                    idx
                )
            except:  # noqa: E722
                raise ValueError(f"Couldn't find worksheet {worksheet}")

        else:
            raise ValueError(f"Couldn't find worksheet index or title {worksheet}")

    def list_worksheets(self, spreadsheet_id):
        """
        Return a list of worksheets in the spreadsheet.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
        `Returns:`
            list
                A List of worksheets order by their index
        """
        worksheets = self.gspread_client.open_by_key(spreadsheet_id).worksheets()
        return [w.title for w in worksheets]

    def get_worksheet_index(self, spreadsheet_id, title):
        """
        Get the first sheet in a Google spreadsheet with the given title. The
        title is case sensitive and the index begins with 0.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            title: str
                The sheet title
        `Returns:`
            str
                The sheet index
        """

        sheets = self.gspread_client.open_by_key(spreadsheet_id).worksheets()
        for index, sheet in enumerate(sheets):
            if sheet.title == title:
                return index
        raise ValueError(f"Couldn't find sheet with title {title}")

    def get_worksheet(self, spreadsheet_id, worksheet=0):
        """
        Create a ``parsons table`` from a sheet in a Google spreadsheet, given the sheet index.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            worksheet: str or int
                The index or the title of the worksheet. The index begins with
                0.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        worksheet = self._get_worksheet(spreadsheet_id, worksheet)
        tbl = Table(worksheet.get_all_values())
        logger.info(f"Retrieved worksheet with {tbl.num_rows} rows.")
        return tbl

    def share_spreadsheet(
        self,
        spreadsheet_id,
        sharee,
        share_type="user",
        role="reader",
        notify=True,
        notify_message=None,
        with_link=False,
    ):
        """
        Share a spreadsheet with a user, group of users, domain and/or the public.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            sharee: str
                User or group e-mail address, domain name to share the spreadsheet
                with. To share publicly, set sharee value to ``None``.
            share_type: str
                The sharee type. Allowed values are: ``user``, ``group``, ``domain``,
                ``anyone``.
            role: str
                The primary role for this user. Allowed values are: ``owner``,
                ``writer``, ``reader``.
            notify: boolean
                Whether to send an email to the target user/domain.
            email_message: str
                The email to be sent if notify kwarg set to True.
            with_link: boolean
                Whether a link is required for this permission.
        """

        spreadsheet = self.gspread_client.open_by_key(spreadsheet_id)
        spreadsheet.share(
            sharee,
            share_type,
            role,
            notify=notify,
            email_message=notify_message,
            with_link=with_link,
        )
        logger.info(f"Shared spreadsheet {spreadsheet_id}.")

    def get_spreadsheet_permissions(self, spreadsheet_id):
        """
        List the permissioned users and groups for a spreadsheet.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        spreadsheet = self.gspread_client.open_by_key(spreadsheet_id)
        tbl = Table(spreadsheet.list_permissions())
        logger.info(f"Retrieved permissions for {spreadsheet_id} spreadsheet.")
        return tbl

    def create_spreadsheet(self, title, editor_email=None, folder_id=None):
        """
        Creates a new Google spreadsheet. Optionally shares the new doc with
        the given email address. Optionally creates the sheet in a specified folder.

        `Args:`
            title: str
                The human-readable title of the new spreadsheet
            editor_email: str (optional)
                Email address which should be given permissions on this spreadsheet.
                Tip: You may want to share this file with the service account.
            folder_id: str (optional)
                ID of the Google folder where the spreadsheet should be created.
                Tip: Get this from the folder URL.
                Anyone shared on the folder will have access to the spreadsheet.

        `Returns:`
            str
                The spreadsheet ID
        """

        spreadsheet = self.gspread_client.create(title, folder_id=folder_id)

        if editor_email:
            self.gspread_client.insert_permission(
                spreadsheet.id,
                editor_email,
                perm_type="user",
                role="writer",
            )

        logger.info(f"Created spreadsheet {spreadsheet.id}")
        return spreadsheet.id

    def delete_spreadsheet(self, spreadsheet_id):
        """
        Deletes a Google spreadsheet.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
        """
        self.gspread_client.del_spreadsheet(spreadsheet_id)
        logger.info(f"Deleted spreadsheet {spreadsheet_id}")

    def add_sheet(self, spreadsheet_id, title=None, rows=100, cols=25):
        """
        Adds a sheet to a Google spreadsheet.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            rows: int
                Number of rows
            cols
                Number of cols

        `Returns:`
            str
                The sheet index
        """
        spreadsheet = self.gspread_client.open_by_key(spreadsheet_id)
        spreadsheet.add_worksheet(title, rows, cols)
        sheet_count = len(spreadsheet.worksheets())
        logger.info("Created worksheet.")
        return sheet_count - 1

    def append_to_sheet(
        self, spreadsheet_id, table, worksheet=0, user_entered_value=False, **kwargs
    ):
        """
        Append data from a Parsons table to a Google sheet. Note that the table's columns are
        ignored, as we'll be keeping whatever header row already exists in the Google sheet.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            table: obj
                Parsons table
            worksheet: str or int
                The index or the title of the worksheet. The index begins with
                0.
            user_entered_value: bool (optional)
                If True, will submit cell values as entered (required for entering formulas).
                Otherwise, values will be entered as strings or numbers only.
        """

        # This is in here to ensure backwards compatibility with previous versions of Parsons.
        if "sheet_index" in kwargs:
            worksheet = kwargs["sheet_index"]
            logger.warning("Argument deprecated. Use worksheet instead.")

        sheet = self._get_worksheet(spreadsheet_id, worksheet)

        # Grab the existing data, so we can figure out where to start adding new data as a batch.
        # TODO Figure out a way to do a batch append without having to read the whole sheet first.
        # Maybe use gspread's low-level batch_update().
        existing_table = self.get_worksheet(spreadsheet_id, worksheet)

        # If the existing sheet is blank, then just overwrite the table.
        if existing_table.num_rows == 0:
            return self.overwrite_sheet(
                spreadsheet_id, table, worksheet, user_entered_value
            )

        cells = []
        for row_num, row in enumerate(table.data):
            for col_num, cell in enumerate(row):
                # Add 2 to allow for the header row, and for google sheets indexing starting at 1
                sheet_row_num = existing_table.num_rows + row_num + 2
                cells.append(gspread.Cell(sheet_row_num, col_num + 1, row[col_num]))

        value_input_option = "RAW"
        if user_entered_value:
            value_input_option = "USER_ENTERED"

        # Update the data in one batch
        sheet.update_cells(cells, value_input_option=value_input_option)
        logger.info(f"Appended {table.num_rows} rows to worksheet.")

    def overwrite_sheet(
        self, spreadsheet_id, table, worksheet=0, user_entered_value=False, **kwargs
    ):
        """
        Replace the data in a Google sheet with a Parsons table, using the table's columns as the
        first row.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            table: obj
                Parsons table
            worksheet: str or int
                The index or the title of the worksheet. The index begins with
                0.
            user_entered_value: bool (optional)
                If True, will submit cell values as entered (required for entering formulas).
                Otherwise, values will be entered as strings or numbers only.
        """

        # This is in here to ensure backwards compatibility with previous versions of Parsons.
        if "sheet_index" in kwargs:
            worksheet = kwargs["sheet_index"]
            logger.warning("Argument deprecated. Use worksheet instead.")

        sheet = self._get_worksheet(spreadsheet_id, worksheet)
        sheet.clear()

        value_input_option = "RAW"
        if user_entered_value:
            value_input_option = "USER_ENTERED"

        # Add header row
        sheet.append_row(table.columns, value_input_option=value_input_option)

        cells = []
        for row_num, row in enumerate(table.data):
            for col_num, cell in enumerate(row):
                # We start at row #2 to keep room for the header row we added above
                cells.append(gspread.Cell(row_num + 2, col_num + 1, row[col_num]))

        # Update the data in one batch
        sheet.update_cells(cells, value_input_option=value_input_option)
        logger.info("Overwrote worksheet.")

    def format_cells(self, spreadsheet_id, range, cell_format, worksheet=0):
        """
        Format the cells of a worksheet.

        `Args:`
            spreadsheet_id: str
                The ID of the spreadsheet (Tip: Get this from the spreadsheet URL)
            range: str
                The cell range to format. E.g. ``"A2"`` or ``"A2:B100"``
            cell_format: dict
                The formatting to apply to the range. Full options are specified in
                the GoogleSheets `API documentation <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#cellformat>`_.
            worksheet: str or int
                The index or the title of the worksheet. The index begins with
                0.

        **Examples**

        .. code-block:: python

            # Set 'A4' cell's text format to bold
            gs.format_cells(sheet_id, "A4", {"textFormat": {"bold": True}}, worksheet=0)

            # Color the background of 'A2:B2' cell range yellow,
            # change horizontal alignment, text color and font size
            gs.format_cells.format(sheet_id, "A2:B2", {
                "backgroundColor": {
                    "red": 0.0,
                    "green": 0.0,
                    "blue": 0.0
                    },
                "horizontalAlignment": "CENTER",
                "textFormat": {
                    "foregroundColor": {
                        "red": 1.0,
                        "green": 1.0,
                        "blue": 0.0
                        },
                        "fontSize": 12,
                        "bold": True
                        }
                    }, worksheet=0)

        """  # noqa: E501,E261

        ws = self._get_worksheet(spreadsheet_id, worksheet)
        ws.format(range, cell_format)
        logger.info("Formatted worksheet")

    def read_sheet(self, spreadsheet_id, sheet_index=0):
        # Deprecated method v0.14 of Parsons.

        logger.warning("Deprecated method. Use get_worksheet() instead.")
        return self.get_worksheet(spreadsheet_id, sheet_index)

    def read_sheet_with_title(self, spreadsheet_id, title):
        # Deprecated method v0.14 of Parsons.

        logger.warning("Deprecated method. Use get_worksheet() instead.")
        return self.get_worksheet(spreadsheet_id, title)

    def get_sheet_index_with_title(self, spreadsheet_id, title):
        # Deprecated method v0.14 of Parsons.

        logger.warning("Deprecated method. Use get_worksheet_index   instead.")
        return self.get_worksheet_index(spreadsheet_id, title)



class AcquisitionTypes:
    """A class to access the Acquisition Types PDI API endpoint."""

    def __init__(self):
        acqtypes_endpoint = "/acquisition_types"
        self.url_acqtypes = self.base_url + acqtypes_endpoint

        super().__init__()

    def get_acquisition_types(self, limit: int = None):
        """Get a list of Acquisition Types.
        `Args:`
            limit: int
                Specify limit to return.

        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        return self._request(self.url_acqtypes, limit=limit)

    def create_acquisition_type(
        self,
        acquisition_type: str,
        acquisition_description: str,
        acquisition_method: str,
        page_default: str = None,
    ):
        """
        Create a new Acquisition Type
        `Args:`
            acquisition_type (string): The acquisition type
            acquisition_description (string): The acquisition description
            acquisition_method (string): The acquisition method
            Options are:
                "Phone"
                "Canvass"
                "Mail"
                "IVR"
                "Text Message"
                "Email"
                "Event"
                "Online"
                "Social"
                "Site"
                "Other Method" ,
            pageDefault (string, optional): The page default.
                "Lookup" (Lookup Page)
                "WalkList" (Create Lists & Files - Walk List)
                "PhoneList" (Create Lists & Files - Phone List)
                "PhoneBank" (Online Phone Bank)
                "Canvassing" (Mobile Canvassing Device)
                "Import" (Imports)
            }
        """
        payload = {
            "acquisitionType": acquisition_type,
            "acquisitionDescription": acquisition_description,
            "acquisitionMethod": acquisition_method,
            "pageDefault": page_default,
        }
        return self._request(self.url_acqtypes, req_type="POST", post_data=payload)

    def get_acquisition_type(self, id: str):
        """
        Get a Acquisition Type by id.
        `Args:`
            id: str
                The Acquisition Type id
        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        return self._request(f"{self.url_acqtypes}/{id}")

    def delete_acquisition_type(self, id: str):
        """
        Delete a Acquisition Type by id.
        `Args:`
            id: str
                The Acquisition Type id
        """
        return self._request(f"{self.url_acqtypes}/{id}", req_type="DELETE")

    def update_acquisition_type(
        self,
        id: str,
        acquisition_type: str,
        acquisition_description: str,
        acquisition_method: str,
        page_default: str = None,
    ):
        """
        Update Acquisition Type
        `Args:`
            acquisition_type (string): The acquisition type
            acquisition_description (string): The acquisition description
            acquisition_method (string): The acquisition method
            Options are:
                "Phone"
                "Canvass"
                "Mail"
                "IVR"
                "Text Message"
                "Email"
                "Event"
                "Online"
                "Social"
                "Site"
                "Other Method" ,
            pageDefault (string, optional): The page default.
                "Lookup" (Lookup Page)
                "WalkList" (Create Lists & Files - Walk List)
                "PhoneList" (Create Lists & Files - Phone List)
                "PhoneBank" (Online Phone Bank)
                "Canvassing" (Mobile Canvassing Device)
                "Import" (Imports)
            }
        """
        payload = {
            "acquisitionType": acquisition_type,
            "acquisitionDescription": acquisition_description,
            "acquisitionMethod": acquisition_method,
            "pageDefault": page_default,
        }
        return self._request(
            f"{self.url_acqtypes}/{id}", req_type="PUT", post_data=payload
        )



import unittest
from test.utils import assert_matching_tables

import requests_mock
from parsons import Table, Zoom

ACCOUNT_ID = "fakeAccountID"
CLIENT_ID = "fakeClientID"
CLIENT_SECRET = "fakeClientSecret"

ZOOM_URI = "https://api.zoom.us/v2/"
ZOOM_AUTH_CALLBACK = "https://zoom.us/oauth/token"


class TestZoom(unittest.TestCase):
    @requests_mock.Mocker()
    def setUp(self, m):
        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        self.zoom = Zoom(ACCOUNT_ID, CLIENT_ID, CLIENT_SECRET)

    @requests_mock.Mocker()
    def test_get_users(self, m):
        user_json = {
            "page_count": 1,
            "page_number": 1,
            "page_size": 30,
            "total_records": 1,
            "users": [
                {
                    "id": "C5A2nRWwTMm_hXyJb1JXMh",
                    "first_name": "Bob",
                    "last_name": "McBob",
                    "email": "bob@bob.com",
                    "type": 2,
                    "pmi": 8374523641,
                    "timezone": "America/New_York",
                    "verified": 1,
                    "dept": "",
                    "created_at": "2017-10-06T15:22:34Z",
                    "last_login_time": "2020-05-06T16:50:45Z",
                    "last_client_version": "",
                    "language": "",
                    "phone_number": "",
                    "status": "active",
                }
            ],
        }

        tbl = Table(
            [
                {
                    "id": "C5A2nRWwTMm_hXyJb1JXMh",
                    "first_name": "Bob",
                    "last_name": "McBob",
                    "email": "bob@bob.com",
                    "type": 2,
                    "pmi": 8374523641,
                    "timezone": "America/New_York",
                    "verified": 1,
                    "dept": "",
                    "created_at": "2017-10-06T15:22:34Z",
                    "last_login_time": "2020-05-06T16:50:45Z",
                    "last_client_version": "",
                    "language": "",
                    "phone_number": "",
                    "status": "active",
                }
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "users", json=user_json)
        assert_matching_tables(self.zoom.get_users(), tbl)

    @requests_mock.Mocker()
    def test_get_meeting_participants(self, m):
        participants = {
            "page_count": 1,
            "page_size": 30,
            "total_records": 4,
            "next_page_token": "",
            "participants": [
                {
                    "id": "",
                    "user_id": "16778240",
                    "name": "Barack Obama",
                    "user_email": "",
                    "join_time": "2020-04-24T21:00:26Z",
                    "leave_time": "2020-04-24T22:24:38Z",
                    "duration": 5052,
                    "attentiveness_score": "",
                },
                {
                    "id": "",
                    "user_id": "16779264",
                    "name": "",
                    "user_email": "",
                    "join_time": "2020-04-24T21:00:45Z",
                    "leave_time": "2020-04-24T22:24:38Z",
                    "duration": 5033,
                    "attentiveness_score": "",
                },
            ],
        }

        tbl = Table(
            [
                {
                    "id": "",
                    "user_id": "16778240",
                    "name": "Barack Obama",
                    "user_email": "",
                    "join_time": "2020-04-24T21:00:26Z",
                    "leave_time": "2020-04-24T22:24:38Z",
                    "duration": 5052,
                    "attentiveness_score": "",
                },
                {
                    "id": "",
                    "user_id": "16779264",
                    "name": "",
                    "user_email": "",
                    "join_time": "2020-04-24T21:00:45Z",
                    "leave_time": "2020-04-24T22:24:38Z",
                    "duration": 5033,
                    "attentiveness_score": "",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "report/meetings/123/participants", json=participants)
        assert_matching_tables(self.zoom.get_past_meeting_participants(123), tbl)

    @requests_mock.Mocker()
    def test_get_meeting_registrants(self, m):
        registrants = {
            "page_count": 1,
            "page_size": 30,
            "total_records": 4,
            "next_page_token": "",
            "registrants": [
                {
                    "id": "",
                    "user_id": "16778240",
                    "name": "Barack Obama",
                    "user_email": "",
                    "purchasing_time_frame": "Within a month",
                    "role_in_purchase_process": "Not involved",
                },
                {
                    "id": "",
                    "user_id": "16779264",
                    "name": "",
                    "user_email": "",
                    "purchasing_time_frame": "Within a month",
                    "role_in_purchase_process": "Not involved",
                },
            ],
        }

        tbl = Table(
            [
                {
                    "id": "",
                    "user_id": "16778240",
                    "name": "Barack Obama",
                    "user_email": "",
                    "purchasing_time_frame": "Within a month",
                    "role_in_purchase_process": "Not involved",
                },
                {
                    "id": "",
                    "user_id": "16779264",
                    "name": "",
                    "user_email": "",
                    "purchasing_time_frame": "Within a month",
                    "role_in_purchase_process": "Not involved",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "meetings/123/registrants", json=registrants)
        assert_matching_tables(self.zoom.get_meeting_registrants(123), tbl)

    @requests_mock.Mocker()
    def test_get_user_webinars(self, m):
        webinars = {
            "page_count": 1,
            "page_size": 30,
            "total_records": 4,
            "next_page_token": "",
            "webinars": [
                {
                    "uuid": "dsghfkhaewfds",
                    "id": "",
                    "host_id": "24654130000000",
                    "topic": "My Webinar",
                    "agenda": "Learn more about Zoom APIs",
                    "type": "5",
                    "duration": "60",
                    "start_time": "2019-09-24T22:00:00Z",
                    "timezone": "America/Los_Angeles",
                    "created_at": "2019-08-30T22:00:00Z",
                    "join_url": "https://zoom.us/0001000/awesomewebinar",
                },
                {
                    "uuid": "dhf8as7dhf",
                    "id": "",
                    "host_id": "24654130000345",
                    "topic": "My Webinar",
                    "agenda": "Learn more about Zoom APIs",
                    "type": "5",
                    "duration": "60",
                    "start_time": "2019-09-24T22:00:00Z",
                    "timezone": "America/Los_Angeles",
                    "created_at": "2019-08-30T22:00:00Z",
                    "join_url": "https://zoom.us/0001000/awesomewebinar",
                },
            ],
        }

        tbl = Table(
            [
                {
                    "uuid": "dsghfkhaewfds",
                    "id": "",
                    "host_id": "24654130000000",
                    "topic": "My Webinar",
                    "agenda": "Learn more about Zoom APIs",
                    "type": "5",
                    "duration": "60",
                    "start_time": "2019-09-24T22:00:00Z",
                    "timezone": "America/Los_Angeles",
                    "created_at": "2019-08-30T22:00:00Z",
                    "join_url": "https://zoom.us/0001000/awesomewebinar",
                },
                {
                    "uuid": "dhf8as7dhf",
                    "id": "",
                    "host_id": "24654130000345",
                    "topic": "My Webinar",
                    "agenda": "Learn more about Zoom APIs",
                    "type": "5",
                    "duration": "60",
                    "start_time": "2019-09-24T22:00:00Z",
                    "timezone": "America/Los_Angeles",
                    "created_at": "2019-08-30T22:00:00Z",
                    "join_url": "https://zoom.us/0001000/awesomewebinar",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "users/123/webinars", json=webinars)
        assert_matching_tables(self.zoom.get_user_webinars(123), tbl)

    @requests_mock.Mocker()
    def test_get_past_webinar_participants(self, m):
        participants = {
            "page_count": 1,
            "page_size": 30,
            "total_records": 4,
            "next_page_token": "",
            "participants": [
                {
                    "id": "",
                    "user_id": "sdfjkldsf87987",
                    "name": "Barack",
                    "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                    "join_time": "2019-02-01T12:34:12.660Z",
                    "leave_time": "2019-03-01T12:34:12.660Z",
                    "duration": "20",
                },
                {
                    "id": "",
                    "user_id": "sdfjkldsfdfgdfg",
                    "name": "Joe",
                    "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                    "join_time": "2019-02-01T12:34:12.660Z",
                    "leave_time": "2019-03-01T12:34:12.660Z",
                    "duration": "20",
                },
            ],
        }

        tbl = Table(
            [
                {
                    "id": "",
                    "user_id": "sdfjkldsf87987",
                    "name": "Barack",
                    "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                    "join_time": "2019-02-01T12:34:12.660Z",
                    "leave_time": "2019-03-01T12:34:12.660Z",
                    "duration": "20",
                },
                {
                    "id": "",
                    "user_id": "sdfjkldsfdfgdfg",
                    "name": "Joe",
                    "user_email": "riya@sdfjkldsf87987.fdjfhdf",
                    "join_time": "2019-02-01T12:34:12.660Z",
                    "leave_time": "2019-03-01T12:34:12.660Z",
                    "duration": "20",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "report/webinars/123/participants", json=participants)
        assert_matching_tables(self.zoom.get_past_webinar_participants(123), tbl)

    @requests_mock.Mocker()
    def test_get_webinar_registrants(self, m):
        registrants = {
            "page_count": 1,
            "page_size": 30,
            "total_records": 4,
            "next_page_token": "",
            "registrants": [
                {
                    "id": "",
                    "email": "barack@obama.com",
                    "first_name": "Barack",
                    "last_name": "Obama",
                    "address": "dsfhkdjsfh st",
                    "city": "jackson heights",
                    "country": "US",
                    "zip": "11371",
                    "state": "NY",
                    "phone": "00000000",
                    "industry": "Food",
                    "org": "Cooking Org",
                    "job_title": "Chef",
                    "purchasing_time_frame": "1-3 months",
                    "role_in_purchase_process": "Influencer",
                    "no_of_employees": "10",
                    "comments": "Looking forward to the Webinar",
                    "custom_questions": [
                        {
                            "title": "What do you hope to learn from this Webinar?",
                            "value": "Look forward to learning how you come up with recipes and services",  # noqa: E501
                        }
                    ],
                    "status": "approved",
                    "create_time": "2019-02-26T23:01:16.899Z",
                    "join_url": "https://zoom.us/webinar/mywebinariscool",
                },
                {
                    "id": "",
                    "email": "joe@biden.com",
                    "first_name": "Joe",
                    "last_name": "Biden",
                    "address": "dsfhkdjsfh st",
                    "city": "jackson heights",
                    "country": "US",
                    "zip": "11371",
                    "state": "NY",
                    "phone": "00000000",
                    "industry": "Food",
                    "org": "Cooking Org",
                    "job_title": "Chef",
                    "purchasing_time_frame": "1-3 months",
                    "role_in_purchase_process": "Influencer",
                    "no_of_employees": "10",
                    "comments": "Looking forward to the Webinar",
                    "custom_questions": [
                        {
                            "title": "What do you hope to learn from this Webinar?",
                            "value": "Look forward to learning how you come up with recipes and services",  # noqa: E501
                        }
                    ],
                    "status": "approved",
                    "create_time": "2019-02-26T23:01:16.899Z",
                    "join_url": "https://zoom.us/webinar/mywebinariscool",
                },
            ],
        }

        tbl = Table(
            [
                {
                    "id": "",
                    "email": "barack@obama.com",
                    "first_name": "Barack",
                    "last_name": "Obama",
                    "address": "dsfhkdjsfh st",
                    "city": "jackson heights",
                    "country": "US",
                    "zip": "11371",
                    "state": "NY",
                    "phone": "00000000",
                    "industry": "Food",
                    "org": "Cooking Org",
                    "job_title": "Chef",
                    "purchasing_time_frame": "1-3 months",
                    "role_in_purchase_process": "Influencer",
                    "no_of_employees": "10",
                    "comments": "Looking forward to the Webinar",
                    "custom_questions": [
                        {
                            "title": "What do you hope to learn from this Webinar?",
                            "value": "Look forward to learning how you come up with recipes and services",  # noqa: E501
                        }
                    ],
                    "status": "approved",
                    "create_time": "2019-02-26T23:01:16.899Z",
                    "join_url": "https://zoom.us/webinar/mywebinariscool",
                },
                {
                    "id": "",
                    "email": "joe@biden.com",
                    "first_name": "Joe",
                    "last_name": "Biden",
                    "address": "dsfhkdjsfh st",
                    "city": "jackson heights",
                    "country": "US",
                    "zip": "11371",
                    "state": "NY",
                    "phone": "00000000",
                    "industry": "Food",
                    "org": "Cooking Org",
                    "job_title": "Chef",
                    "purchasing_time_frame": "1-3 months",
                    "role_in_purchase_process": "Influencer",
                    "no_of_employees": "10",
                    "comments": "Looking forward to the Webinar",
                    "custom_questions": [
                        {
                            "title": "What do you hope to learn from this Webinar?",
                            "value": "Look forward to learning how you come up with recipes and services",  # noqa: E501
                        }
                    ],
                    "status": "approved",
                    "create_time": "2019-02-26T23:01:16.899Z",
                    "join_url": "https://zoom.us/webinar/mywebinariscool",
                },
            ]
        )

        m.post(ZOOM_AUTH_CALLBACK, json={"access_token": "fakeAccessToken"})
        m.get(ZOOM_URI + "webinars/123/registrants", json=registrants)
        assert_matching_tables(self.zoom.get_webinar_registrants(123), tbl)



"""NGPVAN Contact Notes Endpoints"""

from parsons.etl.table import Table
import logging

logger = logging.getLogger(__name__)


class ContactNotes(object):
    def __init__(self, van_connection):
        self.connection = van_connection

    def get_contact_notes(self, van_id):
        """
        Get custom fields.

        `Args:`
            van_id : str
                VAN ID for the person to get notes for.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request(f"people/{van_id}/notes"))
        logger.info(f"Found {tbl.num_rows} custom fields.")
        return tbl

    def create_contact_note(
        self, van_id, text, is_view_restricted, note_category_id=None
    ):
        """
        Create a contact note

        `Args:`
            van_id: str
                VAN ID for the person this note will be applied to.
            text: str
                The content of the note.
            is_view_restricted: bool
                Set to true if the note should be restricted only to certain users within
                the current context; set to false if the note may be viewed by any user
                in the current context.
            note_category_id: int
                Optional; if set, the note category for this note.
        `Returns:`
            int
              The note ID.
        """
        note = {"text": text, "isViewRestricted": is_view_restricted}
        if note_category_id is not None:
            note["category"] = {"noteCategoryId": note_category_id}

        r = self.connection.post_request(f"people/{van_id}/notes", json=note)
        logger.info(f"Contact note {r} created.")
        return r



import logging

logger = logging.getLogger(__name__)


class Events:
    """A class for interacting with PDI events via PDIs API"""

    def __init__(self):
        self.events_url = self.base_url + "/events"
        self.calendars_url = self.base_url + "/calendars"
        self.eventactivities_url = self.base_url + "/eventActivities"
        self.activites_url = self.base_url + "/activities"
        self.activityassignment_url = self.base_url + "/eventActivityAssignments"

        super().__init__()

    def get_events(self, first_event_date: str, last_event_date: str, limit=None):
        """Get a table of PDI events in a given time frame

        `Args:`
            first_event_date: str
                First date in the timeframe from which you want events formatted at 'yyy-MM-dd'
            last_event_date: str
                Last date in the timeframe from which you want events formatted at 'yyy-MM-dd'
            limit: int
                The max number of events to return

        `Returns:`
            parsons.Table
                A Parsons table containing all requested events data.
        """

        params = {
            "startDate": first_event_date,
            "endDate": last_event_date,
        }

        return self._request(self.events_url, args=params, limit=limit)

    def get_event_invitations(self, event_id: str, expand=True, limit=None):
        """Get a table of PDI event invitations for a specified event

        `Args:`
            event_id: str
                ID of event for which you want invitations
            expand: bool
                If True returns columns for contact (and all contact info) and event)

        `Returns:`
            parsons.Table
                A Parsons table containing all requested event invitation data.
        """

        params = {"expand": expand}

        return self._request(
            f"{self.events_url}/{event_id}/invitations", args=params, limit=limit
        )

    def create_event(
        self,
        calendar_id: str,
        location_id: str,
        event_name: str,
        start_datetime: str,
        end_datetime: str,
        description=None,
        all_day=False,
        recurrencetype=None,
        recurrence_end_datetime=None,
        host_phone=None,
        host_email=None,
        website=None,
    ):
        """Create event in a specified calendar

        `Args:`
            calendar_id: str
                The calendar in which you'd like to create an event
            location_id: str
                The unique ID of the PDI location this event took place/is to take place at
            event_name: str
                The name of your event
            description: str
                A short description for your event
            start_datetime: str
                The start datetime of the event in UTC timezone formatted as
                yyyy-MM-ddThh:mm:ss.fffZ
            end_datetime: str
                The end date formatted like start_datetime
            is_all_day: bool
                set to True if event is an all day event. Defaults to False
            recurrencetype: str
                Either 'daily', 'weekly', or 'monthly'. Defaults to None
            recurrence_end_datetime: str
                The end time of the last recurrence of the event formatted as
                yyyy-MM-ddThh:mm:ss.fffZ
            host_phone: str
                An optional contact phone number for the host. Defaults to None
            host_email: str
                An optional contact email for the host. Defaults to None
            website: str
                An optional website for the event. Defualts to None

        `Returns:`
            dict
                Response from PDI in dictionary object

        """

        payload = {
            "locationId": location_id,
            "recurrenceType": recurrencetype,
            "name": event_name,
            "description": description,
            "startDateTimeUtc": start_datetime,
            "endDateTimeUtc": end_datetime,
            "isAllDay": str(all_day).lower(),
            "recurrenceEndDateTimeUtc": recurrence_end_datetime,
            "phone": host_phone,
            "contactEmail": host_email,
            "website": website,
        }

        response = self._request(
            self.calendars_url + f"/{calendar_id}" + "/events",
            req_type="POST",
            post_data=payload,
        )
        event_id = response["id"]
        logger.info(f"Created event {event_name} (id: {event_id})")

        return response

    def create_event_with_activity(
        self,
        calendar_id: str,
        location_id: str,
        activity_id: str,
        event_name: str,
        activity_name: str,
        start_datetime: str,
        end_datetime: str,
        description=None,
        all_day=False,
        recurrencetype=None,
        recurrence_end_datetime=None,
        host_phone=None,
        host_email=None,
        website=None,
        signup_goal=None,
    ):
        """Create event in a specified calendar with an associated activity. The activty will
        be assigned the same start, end time, and recurrance settings as the event.

            `Args:`
                calendar_id: str
                    The unique ID of the calendar in which you'd like to create an event
                location_id: str
                    The unique ID of the PDI location whek this event took place/is to take
                    place
                activity_id:
                    The unique ID of the activity type you'd like to add to the event
                event_name: str
                    The name of your event
                activity_name: str
                    The name of your activity. e.g. 'Pictionary!'
                description: str
                    A short description for your event
                start_datetime: str
                    The start datetime of the event in UTC timezone formatted as
                    yyyy-MM-ddThh:mm:ss.fffZ
                end_datetime: str
                    The end date formatted like start_datetime
                is_all_day = bool
                    set to True if event is an all day event. Defaults to False
                recurrencetype: str
                    Either 'daily', 'weekly', or 'monthly'. Defaults to None
                recurrence_end_datetime: str
                    The end time of the last recurrence of the event formatted as
                    yyyy-MM-ddThh:mm:ss.fffZ
                host_phone: str
                    An optional contact phone number for the host. Defaults to None
                host_email: str
                    An optional contact email for the host. Defaults to None
                website: str
                    An optional website for the event. Defualts to None
                signup_goal: int
                    The goal of how many people you want to complete the activity
            `Returns:`
                dict
                    Response from PDI in dictionary object
        """
        event_data = self.create_event(
            calendar_id,
            location_id,
            event_name,
            start_datetime,
            end_datetime,
            description,
            all_day,
            recurrencetype,
            recurrence_end_datetime,
            host_phone,
            host_email,
            website,
        )
        event_id = event_data["id"]
        logger.info(f"Created event {event_name} (id: {event_id})")

        event_activity_payload = {
            "CalendarId": calendar_id,
            "EventId": event_id,
            "ActivityId": activity_id,
            "LocationId": location_id,
            "RecurrenceType": recurrencetype,
            "Name": activity_name,
            "Description": None,
            "StartDateTimeUtc": start_datetime,
            "EndDateTimeUtc": end_datetime,
            "CountGoal": signup_goal,
            "RecurrenceEndDateTimeUtc": recurrence_end_datetime,
        }

        response = self._request(
            self.eventactivities_url, req_type="POST", post_data=event_activity_payload
        )
        logger.info(
            f"Created activity {activity_name} for event {event_name} (id: {event_id})"
        )

        return response

    def create_event_activity(
        self,
        calendar_id: str,
        event_id: str,
        activity_id: str,
        location_id: str,
        activity_name: str,
        start_datetime: str,
        end_datetime: str,
        description=None,
        recurrencetype=None,
        recurrence_end_datetime=None,
        signup_goal=None,
    ):
        """Create event in a specified calendar with an associated activity

        `Args:`
            calendar_id: str
                The unique ID of the calendar in which you'd like to create an event
            event_id: str
                The unique ID of the event this activity is to be associated with
            activity_id:
                The unique ID of the activity type you'd like to add to the event
            location_id: str
                The unique ID of the PDI location where this event took place/is to take
                place
            activity_name: str
                The name of your activity. e.g. 'Pictionary!'
            description: str
                A short description for your event activity
            start_datetime: str
                The start datetime of the event in UTC timezone formatted as
                yyyy-MM-ddThh:mm:ss.fffZ
            end_datetime: str
                The end date formatted like start_datetime
            recurrencetype: str
                Either 'daily', 'weekly', or 'monthly'. Defaults to None
            recurrence_end_datetime: str
                The end time of the last recurrence of the event formatted as
                yyyy-MM-ddThh:mm:ss.fffZ
            signup_goal: int
                The goal of how many people you want to complete the activity


        `Returns:`
            dict
                Response from PDI in dictionary object
        """

        event_activity_payload = {
            "CalendarId": calendar_id,
            "EventId": event_id,
            "ActivityId": activity_id,
            "LocationId": location_id,
            "RecurrenceType": recurrencetype,
            "Name": activity_name,
            "Description": description,
            "StartDateTimeUtc": start_datetime,
            "EndDateTimeUtc": end_datetime,
            "CountGoal": signup_goal,
            "RecurrenceEndDateTimeUtc": recurrence_end_datetime,
        }

        response = self._request(
            self.eventactivities_url, req_type="POST", post_data=event_activity_payload
        )
        logger.info(f"Created activity {activity_name} for event {event_id})")

        return response

    def create_invitation(
        self,
        event_id: str,
        contact_id: str,
        status: str,
        attended: bool,
        confirmed=False,
        specific_occurrence_start=None,
    ):
        """Create a PDI event invitation indicating a contact has been registered for an event
        `Args:`
            event_id: str
                The ID of the event to write the RSVP to
            contact_id: str
                The ID of the contact to which the invitation belongs
            status: str
                Options are: "Yes", "No", "Maybe", "Scheduled", "Invited", "Cancelled",
                "No-Show", "Completed", and ""
            attended: boolean
                Indicates whether contact attended event
            confirmed: boolean
                Indicates whether invitation confirmed they will attend the event. Defaults to
                False
            specific_occurrence_start: str
                If invitation is for a specific occurrence of a recurring event, then the start
                datetime of the event in UTC formatted as yyyy-MM-ddTHH:mm:ss.fffZ
        `Returns:`
            dict
                Response from PDI in dictionary object
        """

        event_invitation_payload = {
            "contactId": contact_id,
            "rsvpStatus": status,
            "isConfirmed": confirmed,
            "attended": attended,
        }

        if specific_occurrence_start:
            event_invitation_payload[
                "specificOcurrenceStartUtc"
            ] = specific_occurrence_start

        response = self._request(
            self.events_url + f"/{event_id}/invitations",
            req_type="POST",
            post_data=event_invitation_payload,
        )
        return response

    def update_invitation(
        self,
        invitation_id: str,
        event_id: str,
        contact_id: str,
        status=None,
        attended=None,
        confirmed=None,
        specific_occurrence_start=None,
    ):
        """Modify a PDI event invitation
        `Args:`
            invitation_id: str
                The ID of the event invitation
            event_id: str
                The ID of the event that corresponds to the invitation
            contact_id: str
                The ID of the contact to which the invitation belongs
            status: str
                Options are: "Yes", "No", "Maybe", "Scheduled", "Invited", "Cancelled",
                "No-Show", "Completed", and ""
            attended: boolean
                Indicates whether contact attended event
            confirmed: boolean
                Indicates whether invitation confirmed they will attend the event
            specific_occurrence_start: str
                If invitation is for a specific occurrence of a recurring event, then the start
                datetime of the event in UTC formatted as yyyy-MM-ddTHH:mm:ss.fffZ
        `Returns:`
            dict
                Response from PDI in dictionary object
        """

        event_invitation_payload = {"contactId": contact_id}

        if status:
            event_invitation_payload["rsvpStatus"] = status
        if confirmed is not None:
            event_invitation_payload["isConfirmed"] = confirmed
        if attended is not None:
            event_invitation_payload["attended"] = attended
        if specific_occurrence_start:
            event_invitation_payload[
                "specificOcurrenceStartUtc"
            ] = specific_occurrence_start

        response = self._request(
            self.events_url + f"/{event_id}/invitations/{invitation_id}",
            req_type="PUT",
            post_data=event_invitation_payload,
        )
        return response

    def create_activity_assignment(
        self,
        eventactivityid: str,
        contact_id: str,
        status: str,
        completed: bool,
        confirmed=False,
        specific_occurrence_start=None,
    ):
        """Create an activity assignement
        `Args:`
            eventactivityid: str
                The ID of the specific event activity you'd like to assign a contact
            contact_id: str
                The ID of the contact to which the assignment belongs
            status: str
                Options are: "Yes", "No", "Maybe", "Scheduled", "Invited", "Cancelled",
                "No-Show", "Completed", and ""
            completed: boolean
                Indicates whether contact attended event
            confirmed: boolean
                Indicates whether invitation confirmed they will attend the event
            specific_occurrence_start: str
                If invitation is for a specific occurrence of a recurring event, then the start
                datetime of the event in UTC formatted as yyyy-MM-ddTHH:mm:ss.fffZ
        `Returns:`
            dict
                Response from PDI in dictionary object
        """

        assignment_payload = {
            "rsvpStatus": status,
            "isConfirmed": confirmed,
            "isShiftWorked": completed,
            "contactId": contact_id,
            "eventActivityId": eventactivityid,
        }

        if specific_occurrence_start:
            assignment_payload["specificOcurrenceStartUtc"] = specific_occurrence_start

        response = self._request(
            self.activityassignment_url, req_type="POST", post_data=assignment_payload
        )

        return response

    def update_activity_assignment(
        self,
        activityassignementid: str,
        eventactivityid: str,
        contact_id: str,
        status=None,
        completed=None,
        confirmed=None,
        specific_occurrence_start=None,
    ):
        """Create an activity assignement
        `Args:`
            activityassignementid: str
                Id of the specific event activity assignement you want to modify
            eventactivityid: str
                The ID of the specific event activity you'd like to assign a contact
            contact_id: str
                The ID of the contact to which the assignment belongs
            status: str
                Options are: "Yes", "No", "Maybe", "Scheduled", "Invited", "Cancelled",
                "No-Show", "Completed", and ""
            completed: boolean
                Indicates whether contact attended event
            confirmed: boolean
                Indicates whether invitation confirmed they will attend the event
            specific_occurrence_start: str
                If invitation is for a specific occurrence of a recurring event, then the start
                datetime of the event in UTC formatted as yyyy-MM-ddTHH:mm:ss.fffZ
        `Returns:`
            dict
                Response from PDI in dictionary object
        """

        assignment_payload = {
            "contactId": contact_id,
            "eventActivityId": eventactivityid,
        }

        if status:
            assignment_payload["rsvpStatus"] = status
        if confirmed is not None:
            assignment_payload["isConfirmed"] = confirmed
        if completed is not None:
            assignment_payload["isShiftWorked"] = completed
        if specific_occurrence_start:
            assignment_payload["specificOcurrenceStartUtc"] = specific_occurrence_start

        response = self._request(
            self.activityassignment_url + f"/{activityassignementid}",
            req_type="PUT",
            post_data=assignment_payload,
        )

        return response

    def get_event_activity_assignments(self, start_date, end_date, expand, limit=None):
        """
        Get a list of event activity assignments.
        Relevant API docs:
            https://api.bluevote.com/docs/index#/EventActivityAssignments

        `Args`:
            start_date: str
                Earliest records to be returned in the API response
                Per the API docs, use "YYYY-MM-DD" format

            end_date: str
                Latest records to be returned in the API response.
                Per the API docs, use "YYYY-MM-DD" format

            expand: bool
                Parameter to determine if we return the list of shift assigments
                expanded or not

            limit: int
                Specify limit to return (max=2000)

        `Returns`:
            Parsons Table with event activity assignment responses
        """

        if limit and limit > 2000:
            raise ValueError("Maximum allowed limit is 2000")

        params = {"startDate": start_date, "endDate": end_date, "expand": expand}
        return self._request(self.activityassignment_url, args=params, limit=limit)

    def get_event_activities(self, start_date, end_date, limit=None):
        """
        Get a list of event activities.
        Relevant API docs:
            https://api.bluevote.com/docs/index#!/EventActivities/EventActivities_GetAll

        `Args`:
            start_date: str
                Earliest records to be returned in the API response
                Per the API docs, use "YYYY-MM-DD" format

            end_date: str
                Latest records to be returned in the API response.
                Per the API docs, use "YYYY-MM-DD" format

            limit: int
                Specify limit to return (max=2000)

        `Returns`:
            Parsons Table with event activity responses
        """

        if limit and limit > 2000:
            raise ValueError("Maximum allowed limit is 2000")

        params = {"startDate": start_date, "endDate": end_date}
        return self._request(self.eventactivities_url, args=params, limit=limit)

    def get_calendars(self, limit=None):
        """
        Gets a list of calendars.
        Relevant API docs:
            https://api.bluevote.com/docs/index#!/Calendars/Calendars_GetAll

        `Args`:
            limit: int
                Specify limit to return (max=2000)

        `Returns`:
            Parsons Table object with id, name, description, and timeZone records
        """

        if limit and limit > 2000:
            raise ValueError("Maximum allowed limit is 2000")

        return self._request(self.calendars_url, limit=limit)



from parsons.etl.table import Table
import os
import pytest

"""
Use this as a marker before any tests that hit live services. That way they'll only run
if you set the "LIVE_TEST" env var.
Example usage:

@mark_live_test
def test_something():
    service = SomeService()
    ...
"""
mark_live_test = pytest.mark.skipif(
    not os.environ.get("LIVE_TEST"), reason="Skipping because not running live test"
)


# Tests whether a table has the expected structure
def validate_list(expected_keys, table):

    if set(expected_keys) != set(table.columns):
        raise KeyError("Not all expected keys found.")

    return True


def assert_matching_tables(table1, table2, ignore_headers=False):
    if ignore_headers:
        data1 = table1.data
        data2 = table2.data
    else:
        data1 = table1
        data2 = table2

    if isinstance(data1, Table) and isinstance(data2, Table):
        assert data1.num_rows == data2.num_rows

    for r1, r2 in zip(data1, data2):
        # Cast both rows to lists, in case they are different types of collections. Must call
        # .items() on dicts to compare content of collections
        if isinstance(r1, dict):
            r1 = r1.items()
        if isinstance(r2, dict):
            r2 = r2.items()

        assert list(r1) == list(r2)



import unittest
import os
import requests_mock
from parsons import VAN
from test.utils import assert_matching_tables


custom_field = [
    {
        "customFieldId": 157,
        "customFieldParentId": None,
        "customFieldName": "Education level",
        "customFieldGroupId": 52,
        "customFieldGroupName": "Education",
        "customFieldGroupType": "Contacts",
        "customFieldTypeId": "S",
        "isEditable": True,
        "isExportable": False,
        "maxTextboxCharacters": None,
        "availableValues": [
            {"id": 1, "name": "High School diploma", "parentValueId": None},
            {"id": 2, "name": "College degree", "parentValueId": None},
            {"id": 3, "name": "Postgraduate degree", "parentValueId": None},
            {"id": 4, "name": "Doctorate", "parentValueId": None},
        ],
    }
]

custom_field_values = [
    {
        "customFieldId": 157,
        "id": 1,
        "name": "High School diploma",
        "parentValueId": None,
    },
    {"customFieldId": 157, "id": 2, "name": "College degree", "parentValueId": None},
    {
        "customFieldId": 157,
        "id": 3,
        "name": "Postgraduate degree",
        "parentValueId": None,
    },
    {"customFieldId": 157, "id": 4, "name": "Doctorate", "parentValueId": None},
]

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestCustomFields(unittest.TestCase):
    def setUp(self):

        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters")

    @requests_mock.Mocker()
    def test_get_custom_fields(self, m):

        m.get(self.van.connection.uri + "customFields", json=custom_field)
        assert_matching_tables(custom_field, self.van.get_custom_fields())

    @requests_mock.Mocker()
    def test_get_custom_field_values(self, m):

        m.get(self.van.connection.uri + "customFields", json=custom_field)
        assert_matching_tables(custom_field_values, self.van.get_custom_fields_values())

    @requests_mock.Mocker()
    def test_get_custom_field(self, m):

        m.get(self.van.connection.uri + "customFields/157", json=custom_field)
        assert_matching_tables(custom_field, self.van.get_custom_field(157))



from abc import ABC, abstractmethod
from typing import Optional
from parsons.etl.table import Table


class DatabaseConnector(ABC):
    """
    An abstract base class that provides a uniform interface for all Parsons database connectors.
    This class should be used in functions instead of the specific database connector classes
    when the functions don't rely on database-specific functionality.

    It ensures that any class that inherits from it implements the methods that are uniform
    operations when working with databases.

    Should you use `DatabaseConnector` instead of `Redshift`/`BigQuery`/etc?

    Overall this class is mostly useful for code in the Parsons library, not code using it.
    There could be some exceptions. In general though, if you are writing a script to do a task
    like moving data out of an API service and into a data warehouse, you probably do not need
    to use DatabaseConnector. You can probably just use the Parsons class that directly corresponds
    with the database that you use.

    Here are more examples of situations where you may or may not need to use DatabaseConnector:

        1. You do not use type annotations, or you don't know what "type annotations" are - No

            If you do not use type annotations for your code, then you do not need to think about
            `DatabaseConnector` when writing your code. This is the most common case. If none
            of the cases below apply to you, then you probably don't need it.

            In this simple example, we are not using type annotations in our code. We don't need
            to think about exactly what class is being passed in. Python will figure it out.

            ```python
            def my_database_function(db):
                    some_data = get_some_data()
                    db.copy("some_table", some_data)

            # These will all just work:
            my_database_function(Redshift())
            my_database_function(MySQL())
            my_database_functon(BigQuery())
            ```

        2. You only use one database in your work - No

            This is where most people will fall. Usually code is not intended to run on
            multiple databases without modification. For example, if you are working for
            an organization that uses Amazon Redshift as your data warehouse, you do not
            need to use `DatabaseConnector` to write ETL scripts to load data into your
            Redshift. It is rare that organizations switch databases. In the cases where
            that does occur, usually more work is required to migrate your environment and
            your vendor-specific SQL than would be saved by using `DatabaseConnector`.

        3. You are writing a sample script or a tutorial - Yes

            If you are using Parsons to write a sample script or tutorial, you should use
            `DatabaseConnector`! If you use `DatabaseConnector` type annotations and the
            `discover_database` function, then your sample code will run on any system.
            This makes it much easier for new programmers to get your code working on
            their system.

        4. Utility code inside Parsons or other libraries - Yes

            If you are writing a utility script inside Parsons or another library meant
            for broad distribution, you should probably use `DatabaseConnector` type
            annotations. This will ensure that your library code will be usable by the
            widest possible set of users, not just users on one specific database.

    Developer Notes:
        This class is an Abstract Base Class (ABC). It's designed to ensure that all classes
        inheriting from it implement certain methods, enforcing a consistent interface across
        database connectors.

        If you need to add a new method to the database connectors, there are three options:
        1. Add the method to this ABC and implement it for all databases.
        2. Add the method to this ABC and implement it for some databases while adding stubs for
           others.
        3. Implement the method on a specific database connector without touching the ABC.

        If you go the second route, you can add a stub method like this:

            .. code-block:: python

                def new_method(self, arg1, arg2):
                    raise NotImplementedError("Method not implemented for this database connector.")
            ```

        This communicates clearly to users that the method does not exist for certain connectors.

        If you go the third route, remember that you're responsible for making sure your new
        method matches the existing methods in other database connectors. For example, if you're
        adding a method that already exists in another connector, like Redshift, you need to ensure
        your new method behaves the same way and has the same parameters with the same types in the
        same order. See the note below for more detail.

    Note:
        The Python type system (as of 3.10.6) will not stop you from breaking the type contract
        of method signatures when implementing a subclass. It is up to the author of a database
        connector to ensure that it satisfies this interface. Be careful to, for example, not
        change the types of the parameters or leave out optional parameters that are specified
        in the interface.

        Any such inconsistencies can cause unexpected runtime errors that will not be caught by
        the type checker.

        It is safe to add additional features to subclasses, such as new methods or extra *optional*
        parameters to specified methods. In general adding new methods is safe, but adding optional
        parameters to methods specified in the interface should be considered bad practice, because
        it could result in unexpected behavior.

    Example usage:

    .. code-block:: python

        def my_function(db: DatabaseConnector, data: Table):
            # Your code here, using the db object

        # Pass an instance of a class that inherits from DatabaseConnector, e.g. Redshift
        my_function(some_db_instance, some_data)

    """

    @abstractmethod
    def table_exists(self, table_name: str) -> bool:
        """Check if a table or view exists in the database.

        `Args:`
            table_name: str
                The table name and schema (e.g. ``myschema.mytable``).

        `Returns:`
            boolean
                ``True`` if the table exists and ``False`` if it does not.
        """
        pass

    @abstractmethod
    def copy(self, tbl: Table, table_name: str, if_exists: str):
        """Copy a :ref:`parsons-table` to the database.

        `Args`:
            tbl (Table):
                Table containing the data to save.
            table_name (str):
                The destination table name (ex. ``my_schema.my_table``).
            if_exists (str):
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
        """
        pass

    @abstractmethod
    def query(self, sql: str, parameters: Optional[list] = None) -> Optional[Table]:
        """Execute a query against the database. Will return ``None`` if the query returns empty.

        To include python variables in your query, it is recommended to pass them as parameters,
        following the `psycopg style
          <http://initd.org/psycopg/docs/usage.html#passing-parameters-to-sql-queries>`.
        Using the ``parameters`` argument ensures that values are escaped properly, and avoids SQL
        injection attacks.

        **Parameter Examples**

        .. code-block:: python

            # Note that the name contains a quote, which could break your query if not escaped
            # properly.
            name = "Beatrice O'Brady"
            sql = "SELECT * FROM my_table WHERE name = %s"
            db.query(sql, parameters=[name])

        .. code-block:: python

            names = ["Allen Smith", "Beatrice O'Brady", "Cathy Thompson"]
            placeholders = ', '.join('%s' for item in names)
            sql = f"SELECT * FROM my_table WHERE name IN ({placeholders})"
            db.query(sql, parameters=names)

        `Args:`
            sql: str
                A valid SQL statement
            parameters: Optional[list]
                A list of python variables to be converted into SQL values in your query

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        pass



from parsons.databases.database.database import DatabaseCreateStatement
import parsons.databases.postgres.constants as consts

import petl
import logging

logger = logging.getLogger(__name__)


class PostgresCreateStatement(DatabaseCreateStatement):
    def __init__(self):
        super().__init__()

        self.COL_NAME_MAX_LEN = consts.COL_NAME_MAX_LEN
        self.REPLACE_CHARS = consts.REPLACE_CHARS

        # Postgres doesn't have a medium int
        self.MEDIUMINT = self.INT

        # Currently py floats are coded as Postgres decimals
        self.FLOAT = consts.DECIMAL

        # Max length of a Redshift VARCHAR column
        self.VARCHAR_MAX = consts.VARCHAR_MAX

        # List of varchar lengths to use for columns -- this list needs to be in order from
        # smallest to largest
        self.VARCHAR_STEPS = consts.VARCHAR_STEPS

    # the default behavior is f"{col}_"
    def _rename_reserved_word(self, col, index):
        """Return the renamed column.

        `Args`:
            col: str
                The column to rename.
            index: int
                (Optional) The index of the column.
        `Returns`:
            str
                The rename column.
        """
        return f"col_{index}"

    def create_statement(
        self,
        tbl,
        table_name,
        padding=None,
        distkey=None,
        sortkey=None,
        varchar_max=None,
        varchar_truncate=True,
        columntypes=None,
        strict_length=True,
    ):
        # Generate a table create statement. Distkeys and sortkeys are only used by
        # Redshift and should not be passed when generating a create statement for
        # Postgres.

        if tbl.num_rows == 0:
            raise ValueError("Table is empty. Must have 1 or more rows.")

        # Validate and rename column names if needed
        tbl.table = petl.setheader(tbl.table, self.column_name_validate(tbl.columns))

        mapping = self.generate_data_types(tbl)

        if padding:
            mapping["longest"] = self.vc_padding(mapping, padding)
        elif not strict_length:
            mapping["longest"] = self.vc_step(mapping)

        if varchar_max:
            mapping["longest"] = self.vc_max(mapping, varchar_max)

        if varchar_truncate:
            mapping["longest"] = self.vc_trunc(mapping)

        mapping["longest"] = self.vc_validate(mapping)

        # Add any provided column type overrides
        if columntypes:
            for i in range(len(mapping["headers"])):
                col = mapping["headers"][i]
                if columntypes.get(col):
                    mapping["type_list"][i] = columntypes[col]

        # Enclose in quotes
        mapping["headers"] = [f'"{h}"' for h in mapping["headers"]]

        return self.create_sql(table_name, mapping, distkey=distkey, sortkey=sortkey)

    # This is for backwards compatability
    def data_type(self, val, current_type):
        return self.detect_data_type(val, current_type)

    # This is for backwards compatability
    def is_valid_integer(self, val):
        return self.is_valid_sql_num(val)

    def generate_data_types(self, table):
        # Generate column data types

        longest, type_list = [], []

        cont = petl.records(table.table)

        # Populate empty values for the columns
        for col in table.columns:
            longest.append(0)
            type_list.append("")

        for row in cont:
            for i in range(len(row)):
                # NA is the csv null value
                if type_list[i] == "varchar" or row[i] in ["NA", ""]:
                    pass
                else:
                    var_type = self.data_type(row[i], type_list[i])
                    type_list[i] = var_type

                # Calculate width
                width = len(str(row[i]).encode("utf-8"))
                if width > longest[i]:
                    longest[i] = width

        # In L140 'NA' and '' will be skipped
        # If the entire column is either one of those (or a mix of the two)
        # the type will be empty.
        # Fill with a default varchar
        type_list = [typ or "varchar" for typ in type_list]

        return {"longest": longest, "headers": table.columns, "type_list": type_list}

    def vc_padding(self, mapping, padding):
        # Pad the width of a varchar column

        return [int(c + (c * padding)) for c in mapping["longest"]]

    def vc_step(self, mapping):
        return [self.round_longest(c) for c in mapping["longest"]]

    def vc_max(self, mapping, columns):
        # Set the varchar width of a column to the maximum

        for c in columns:

            try:
                idx = mapping["headers"].index(c)
                mapping["longest"][idx] = self.VARCHAR_MAX

            except KeyError as error:
                logger.error("Could not find column name provided.")
                raise error

        return mapping["longest"]

    def vc_trunc(self, mapping):

        return [
            self.VARCHAR_MAX if c > self.VARCHAR_MAX else c for c in mapping["longest"]
        ]

    def vc_validate(self, mapping):

        return [1 if c == 0 else c for c in mapping["longest"]]

    def create_sql(self, table_name, mapping, distkey=None, sortkey=None):
        # Generate the sql to create the table

        statement = f"create table {table_name} ("

        for i in range(len(mapping["headers"])):
            if mapping["type_list"][i] == "varchar":
                statement = (statement + "\n  {} varchar({}),").format(
                    str(mapping["headers"][i]).lower(), str(mapping["longest"][i])
                )
            else:
                statement = (statement + "\n  " + "{} {}" + ",").format(
                    str(mapping["headers"][i]).lower(), mapping["type_list"][i]
                )

        statement = statement[:-1] + ") "

        if distkey:
            statement += f"\ndistkey({distkey}) "

        if sortkey:
            statement += f"\nsortkey({sortkey})"

        statement += ";"

        return statement

    # This is for backwards compatability
    def column_name_validate(self, columns):
        return self.format_columns(columns, col_prefix="col_")

    @staticmethod
    def round_longest(longest):
        # Find the value that will work best to fit our longest column value
        for step in consts.VARCHAR_STEPS:
            # Make sure we have padding
            if longest < step / 2:
                return step

        return consts.VARCHAR_MAX



from functools import wraps
import paramiko


def connection_exists(args, kwargs):
    if any([isinstance(arg, paramiko.sftp_client.SFTPClient) for arg in args]):
        return True
    if "connection" in kwargs and kwargs["connection"]:
        return True
    return False


def connect(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not connection_exists(args, kwargs):
            with args[0].create_connection() as connection:
                kwargs["connection"] = connection
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    return wrapper



"""**TargetSmart Automation**

Parsons provides methods for interacting with TargetSmart Automation Workflows,
a solution for executing custom file processing workflows programmatically. In
some cases, TargetSmart will provide custom list matching solutions using
Automation Workflows. Most TargetSmart clients do not have these workflows and
will only use the Developer API.

**TargetSmart Developer API versus Automation**

TargetSmart's Developer API provides an HTTP-based interface for consuming the
general web services that TargetSmart provides. The TargetSmart Automation
system solely provides a solution for consuming customized file processing
workflows that are provisioned for specific client needs. TargetSmart
Automation is based on SFTP instead of HTTP.

https://docs.targetsmart.com/my_tsmart/automation/developer.html

For most list matching applications, TargetSmart SmartMatch is now the recommended
solution. See `TargetSmartAPI.smartmatch`.

"""

from parsons.sftp.sftp import SFTP
from parsons.etl.table import Table
from parsons.utilities.files import create_temp_file
from parsons.utilities import check_env
import xml.etree.ElementTree as ET
import uuid
import time
import logging
import xmltodict


TS_STFP_HOST = "transfer.targetsmart.com"
TS_SFTP_PORT = 22
TS_SFTP_DIR = "automation"

logger = logging.getLogger(__name__)


# Automation matching documentation can be found here:
# https://docs.targetsmart.com/my_tsmart/automation/developer.html.
class TargetSmartAutomation(object):
    """
    * `Automation overview <https://docs.targetsmart.com/my_tsmart/automation/overview.html>`_
    * `Automation integration doc <https://docs.targetsmart.com/my_tsmart/automation/developer.html>`_
    """  # noqa

    def __init__(self, sftp_username=None, sftp_password=None):

        self.sftp_host = TS_STFP_HOST
        self.sftp_port = TS_SFTP_PORT
        self.sftp_dir = TS_SFTP_DIR
        self.sftp_username = check_env.check("TS_SFTP_USERNAME", sftp_username)
        self.sftp_password = check_env.check("TS_SFTP_PASSWORD", sftp_password)
        self.sftp = SFTP(
            self.sftp_host,
            self.sftp_username,
            self.sftp_password,
            self.sftp_port,
        )

    def match(
        self,
        table,
        job_type,
        job_name=None,
        emails=None,
        call_back=None,
        remove_files=True,
    ):
        """Submit a file for custom data processing using the TargetSmart Automation workflow solution.

        .. warning::
            Table Columns

              Each Automation workflow expects an input file that meets the
              layout requirements provided by TargetSmart. The number of columns
              and column order is significant. So, if it expected 10 columns and
              you only provide 9, it will fail. However, if you provide 10
              columns that are out of order, the job may succeed, but with
              non-optimal results. You can obtain the layout requirements and
              other information about a workflow by visiting the Automation
              console in My TargetSmart. Contact `TargetSmart Client Services
              <mailto:support@targetsmart.com>`_ for support.

        Args:
            table: Parsons Table Object
                A table object with the required columns. Each workflow type
                requires the input file to meet the requirements provided by
                TargetSmart. You can locate the input and output layouts for
                your available workflows using the My TargetSmart Automation
                console.
            job_type: str
                The workflow name to execute. **This is case sensitive.**. You
                can locate the workflow names and other information by visiting
                the Automation console in My TargetSmart.
            job_name: str
                Optional job execution name.
            emails: list
                A list of emails that will received status notifications. This
                is useful in debugging failed jobs.
            call_back: str
                A callback url to which the status will be posted. See
                `TargetSmart documentation <https://docs.targetsmart.com/my_tsmart/automation/developer.html>`_
                for more details.
            remove_files: boolean
                Remove the configuration, file to be matched and matched file from
                the TargetSmart SFTP upon completion or failure of match.

        """  # noqa: E501,E261

        # Generate a match job
        job_name = job_name or str(uuid.uuid1())

        try:
            # Upload table
            self.sftp.put_file(table.to_csv(), f"{self.sftp_dir}/{job_name}_input.csv")
            logger.info(f"Table with {table.num_rows} rows uploaded to TargetSmart.")

            # Create/upload XML configuration
            xml = self.create_job_xml(
                job_type,
                job_name,
                emails=emails,
                status_key=job_name,
                call_back=call_back,
            )
            self.sftp.put_file(xml, f"{self.sftp_dir}/{job_name}.job.xml")
            logger.info(
                f"Payload uploaded to TargetSmart. Job type: {job_type}. Job name: {job_name}"
            )

            # Check xml configuration status
            self.poll_config_status(job_name)

            # Check the status of the match
            self.match_status(job_name)

            # Download the resulting file
            tbl = Table.from_csv(
                self.sftp.get_file(f"{self.sftp_dir}/{job_name}_output.csv")
            )

        finally:
            # Clean up files
            if remove_files:
                self.remove_files(job_name)

        # Return file as a Table
        return tbl

    def execute(self, *args, **kwargs):
        """Most Automation workflows perform list matching. However, it is possible that
        a custom workflow might be provisioned for a client for other types of
        file processing. The ``execute`` method is provided as an alias for the
        ``match`` method which may be a confusing name in these cases.
        """
        self.match(*args, **kwargs)

    def create_job_xml(
        self, job_type, job_name, emails=None, status_key=None, call_back=None
    ):
        # Internal method to create a valid job xml

        job = ET.Element("job")

        # Generate Base XML
        input_file = ET.SubElement(job, "inputfile")
        input_file.text = job_name + "_input.csv"
        output_file = ET.SubElement(job, "outputfile")
        output_file.text = job_name + "_output.csv"
        jobtype = ET.SubElement(job, "jobtype", text=job_type)
        jobtype.text = job_type

        # Add status key
        args = ET.SubElement(job, "args")
        statuskey = ET.SubElement(args, "arg", name="__status_key")
        statuskey.text = status_key or job_name

        # Option args
        if call_back:
            callback = ET.SubElement(args, "arg", name="__http_callback")
            callback.text = call_back

        if emails:
            emails_el = ET.SubElement(args, "arg", name="__emails")
            emails_el.text = ",".join(emails)

        # Write xml to file object
        local_path = create_temp_file(suffix=".xml")
        tree = ET.ElementTree(job)
        tree.write(local_path)
        return local_path

    def poll_config_status(self, job_name, polling_interval=20):
        #  Poll the configuration status

        while True:

            time.sleep(polling_interval)
            if self.config_status(job_name):
                return True
            logger.info(f"Waiting on {job_name} job configuration...")

    def config_status(self, job_name):
        # Check the status of the configuration by parsing the
        # the files in the SFTP directory.

        for f in self.sftp.list_directory(remote_path=self.sftp_dir):

            if f == f"{job_name}.job.xml.good":
                logger.info(f"Match job {job_name} configured.")
                return True

            elif f == f"{job_name}.job.xml.bad":
                logger.info(f"Match job {job_name} configuration error.")
                #  To Do: Lift up the configuration error.
                raise ValueError(
                    "Job configuration failed. If you provided an email"
                    "address, you will be sent more details."
                )

            else:
                pass

        return False

    def match_status(self, job_name, polling_interval=60):
        # You could also poll their API for the status, which was what the original
        # version of the automation matching did. Note: The polling API is public
        # and does expose some metadata. This happens regardless of anything that
        # we do. However, the actually data is only exposed on the secure SFTP.

        while True:

            logger.debug("Match running...")
            for file_name in self.sftp.list_directory(remote_path=self.sftp_dir):

                if file_name == f"{job_name}.finish.xml":

                    xml_file = self.sftp.get_file(
                        f"{self.sftp_dir}/{job_name}.finish.xml"
                    )
                    with open(xml_file, "rb") as x:
                        xml = xmltodict.parse(x, dict_constructor=dict)

                    if xml["jobcontext"]["state"] == "error":
                        # To Do: Parse these in a pretty way
                        logger.info(f"Match Error: {xml['jobcontext']['errors']}")
                        raise ValueError(
                            f"Match job failed. {xml['jobcontext']['errors']}"
                        )

                    elif xml["jobcontext"]["state"] == "success":
                        logger.info("Match complete.")

                        return True

            time.sleep(polling_interval)

    def remove_files(self, job_name):
        # Remove all of the files for the match.

        for file_name in self.sftp.list_directory(remote_path=self.sftp_dir):
            if job_name in file_name:
                self.sftp.remove_file(f"{self.sftp_dir}/{file_name}")
                logger.info(f"{file_name} removed from SFTP.")



import logging
import json

logger = logging.getLogger(__name__)


class BluelinkPerson(object):
    """
    Instantiate BluelinkPerson Class.
    Used for to upserting via Bluelink connector.
    See: https://bluelinkdata.github.io/docs/BluelinkApiGuide#person-object

    `Args:`
        identifiers: list[BluelinkIdentifier]
            A list of BluelinkIdentifier objects.
            A BluelinkPerson must have at least 1 identifier.
        given_name: str
            First name / given name.
        family_name: str
            Last name / family name.
        phones: list[BluelinkPhone]
            A list of BluelinkPhone objects representing phone numbers.
        emails: list[BluelinkEmail]
            A list of BluelinkEmail objects representing email addresses.
        addresses: list[BluelinkAddress]
            A list of BluelinkAddress objects representing postal addresses.
        tags: list[BluelinkTag]
            Simple tags that apply to the person, eg DONOR.
        employer: str
            Name of the persons employer.
        employer_address: BluelinkAddress
            BluelinkAddress of the persons employer.
        occupation: str
            Occupation.
        scores: list[BluelinkScore]
            List of BluelinkScore objects. Scores are numeric scores, ie partisanship model.
        birthdate: str
            ISO 8601 formatted birth date: 'YYYY-MM-DD'
        details: dict
            additional custom data. must be json serializable.
    """

    def __init__(
        self,
        identifiers,
        given_name=None,
        family_name=None,
        phones=None,
        emails=None,
        addresses=None,
        tags=None,
        employer=None,
        employer_address=None,
        occupation=None,
        scores=None,
        birthdate=None,
        details=None,
    ):

        if not identifiers:
            raise Exception(
                "BluelinkPerson requires list of BluelinkIdentifiers with "
                "at least 1 BluelinkIdentifier"
            )

        self.identifiers = identifiers
        self.addresses = addresses
        self.emails = emails
        self.phones = phones
        self.tags = tags
        self.scores = scores

        self.given_name = given_name
        self.family_name = family_name

        self.employer = employer
        self.employer_address = employer_address
        self.occupation = occupation
        self.birthdate = birthdate
        self.details = details

    def __json__(self):
        """The json str representation of this BluelinkPerson object"""
        return json.dumps(self, default=lambda obj: obj.__dict__)

    def __eq__(self, other):
        """A quick and dirty equality check"""
        dself = json.loads(self.__json__())
        dother = json.loads(other.__json__())
        return dself == dother

    def __repr__(self):
        return self.__json__()

    @staticmethod
    def from_table(tbl, dict_to_person):
        """
        Return a list of BluelinkPerson objects from a Parsons Table.

        `Args:`
            tbl: Table
                A parsons Table.
            dict_to_person: Callable[[dict],BluelinkPerson]
                A function that takes a dictionary representation of a table row,
                and returns a BluelinkPerson.
        `Returns:`
           list[BluelinkPerson]
           A list of BluelinkPerson objects.
        """
        return [dict_to_person(row) for row in tbl]


class BluelinkIdentifier(object):
    """
    Instantiate an BluelinkIdentifier object.
    BluelinkIdentifier is necessary for updating BluelinkPerson records.

    `Args:`
        source: str
            External system to which this ID belongs, e.g., “VAN:myCampaign”.
            Bluelink has standardized strings for source. Using these will
            allow Bluelink to correctly understand the external IDs you add.
            source (unlike identifier) is case insensitive.
            examples: BLUELINK, PDI, SALESFORCE, VAN:myCampaign, VAN:myVoters
        identifier: str
            Case-sensitive ID in the external system.
        details: dict
            dictionary of custom fields. must be serializable to json.
    """

    def __init__(self, source, identifier, details=None):
        self.source = source
        self.identifier = identifier
        self.details = details


class BluelinkEmail(object):
    """
    Instantiate an BluelinkEmail object.

    `Args:`
        address: str
            An email address. ie "user@example.com"
        primary: bool
            True if this is known to be the primary email.
        type: str
            Type, eg: "personal", "work"
        status: str
            One of "Potential", "Subscribed", "Unsubscribed", "Bouncing", or "Spam Complaints"
    """

    def __init__(self, address, primary=None, type=None, status=None):
        self.address = address
        self.primary = primary
        self.type = type
        self.status = status


class BluelinkAddress(object):
    """
    Instantiate an BluelinkAddress object.

    `Args`:
        address_lines: list[str]
            A list of street address lines.
        city: str
            City or other locality.
        state: str
            State in ISO 3166-2.
        postal_code: str
            Zip or other postal code.
        country: str
            ISO 3166-1 Alpha-2 country code.
        type: str
            The type. ie: "home", "mailing".
        venue: str
            The venue name, if relevant.
        status: str
            A value representing the status of the address. "Potential", "Verified" or "Bad"
    """

    def __init__(
        self,
        address_lines=None,
        city=None,
        state=None,
        postal_code=None,
        country=None,
        type=None,
        venue=None,
        status=None,
    ):

        self.address_lines = address_lines or []
        self.city = city
        self.state = state
        self.postal_code = postal_code
        self.country = country

        self.type = type
        self.venue = venue
        self.status = status


class BluelinkPhone(object):
    """
    Instantiate a BluelinkPhone object.

    `Args:`
        number: str
            A phone number. May or may not include country code.
        primary: bool
            True if this is known to be the primary phone.
        description: str
            Free for description.
        type: str
            Type, eg: "Home", "Work", "Mobile"
        country: str
            ISO 3166-1 Alpha-2 country code.
        sms_capable: bool
            True if this number can accept SMS.
        do_not_call: bool
            True if this number is on the US FCC Do Not Call Registry.
        details: dict
            Additional data dictionary. Must be json serializable.
    """

    def __init__(
        self,
        number,
        primary=None,
        description=None,
        type=None,
        country=None,
        sms_capable=None,
        do_not_call=None,
        details=None,
    ):
        self.number = number
        self.primary = primary
        self.description = description
        self.type = type
        self.country = country
        self.sms_capable = sms_capable
        self.do_not_call = do_not_call
        self.details = details


class BluelinkTag(object):
    """
    Instantiate a BluelinkTag object.

    `Args:`
        tag: str
            A tag string; convention is either a simple string
            or a string with a prefix separated by a colon, e.g., “DONOR:GRASSROOTS”
    """

    def __init__(self, tag):
        self.tag = tag


class BluelinkScore(object):
    """
    Instantiate a score object.
    Represents some kind of numeric score.

    `Args`:
        score: float
            Numeric score.
        score_type: str
            Type, eg: "Partisanship model".
        source: str
            Original source of this score.
    """

    def __init__(self, score, score_type, source):
        self.score = score
        self.score_type = score_type
        self.source = source



from test.utils import mark_live_test
from parsons import PDI

import pytest
import os


#
# Fixtures and constants
#
def remove_from_env(*env_vars):
    for var in env_vars:
        try:
            del os.environ[var]
        except KeyError:
            pass


#
# Tests
#

# Need to provide environment variables
# PDI_USERNAME, PDI_PASSWORD, PDI_API_TOKEN
@mark_live_test
def test_connection():
    PDI(qa_url=True)


@pytest.mark.parametrize(
    ["username", "password", "api_token"],
    [
        (None, None, None),
        (None, "pass", "token"),
        ("user", None, "token"),
        ("user", "pass", None),
    ],
)
def test_init_error(username, password, api_token):
    remove_from_env("PDI_USERNAME", "PDI_PASSWORD", "PDI_API_TOKEN")
    with pytest.raises(KeyError):
        PDI(username, password, api_token)


@pytest.mark.parametrize(
    ["obj", "exp_obj"],
    [
        ({"a": "a", "b": None, "c": "c"}, {"a": "a", "c": "c"}),
        (
            [{"a": "a", "b": None, "c": "c"}, {"a": "a", "c": None}],
            [{"a": "a", "c": "c"}, {"a": "a"}],
        ),
        ("string", "string"),
    ],
)
def test_clean_dict(mock_pdi, obj, exp_obj):
    assert mock_pdi._clean_dict(obj) == exp_obj



ENV_PARAMETERS = {
    "BLOOMERANG_API_KEY": "env_api_key",
    "BLOOMERANG_CLIENT_ID": "env_client_id",
    "BLOOMERANG_CLIENT_SECRET": "env_client_secret",
}

ID = 123

TEST_DELETE = {"Id": 0, "Type": "string", "Deleted": "true"}

TEST_CREATE_CONSTITUENT = {
    "Type": "Individual",
    "Status": "Active",
    "FirstName": "string",
    "LastName": "string",
    "MiddleName": "string",
    "Prefix": "string",
    "Suffix": "string",
    "FullName": "string",
    "InformalName": "string",
    "FormalName": "string",
    "EnvelopeName": "string",
    "RecognitionName": "string",
    "JobTitle": "string",
    "Employer": "string",
    "Website": "string",
    "FacebookId": "string",
    "TwitterId": "string",
    "LinkedInId": "string",
    "Gender": "Male",
    "Birthdate": "2020-08-24",
    "ProfilePictureType": "None",
    "PrimaryEmail": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Value": "user@example.com",
        "IsPrimary": "true",
        "IsBad": "true",
    },
    "PrimaryPhone": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Extension": "string",
        "Number": "string",
        "IsPrimary": "true",
    },
    "PrimaryAddress": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Street": "string",
        "City": "string",
        "State": "string",
        "PostalCode": "string",
        "Country": "string",
        "IsPrimary": "true",
        "IsBad": "true",
    },
}

TEST_GET_CONSTITUENT = {
    "Id": 0,
    "AccountNumber": 0,
    "IsInHousehold": "true",
    "IsHeadOfHousehold": "true",
    "IsFavorite": "true",
    "FullCustomProfileImageId": 0,
    "FullCustomProfileImageUrl": "string",
    "CroppedCustomProfileImageId": 0,
    "CroppedCustomProfileImageUrl": "string",
    "Type": "Individual",
    "Status": "Active",
    "FirstName": "string",
    "LastName": "string",
    "MiddleName": "string",
    "Prefix": "string",
    "Suffix": "string",
    "FullName": "string",
    "InformalName": "string",
    "FormalName": "string",
    "EnvelopeName": "string",
    "RecognitionName": "string",
    "JobTitle": "string",
    "Employer": "string",
    "Website": "string",
    "FacebookId": "string",
    "TwitterId": "string",
    "LinkedInId": "string",
    "Gender": "Male",
    "Birthdate": "2020-09-08",
    "ProfilePictureType": "None",
    "PrimaryEmail": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Value": "user@example.com",
        "IsPrimary": "true",
        "IsBad": "true",
    },
    "PrimaryPhone": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Extension": "string",
        "Number": "string",
        "IsPrimary": "true",
    },
    "HouseholdId": 0,
    "PreferredCommunicationChannel": "Email",
    "CommunicationRestrictions": ["DoNotCall"],
    "CommunicationRestrictionsUpdateReason": "string",
    "EmailInterestType": "All",
    "CustomEmailInterests": [{"Id": 0, "Name": "string"}],
    "EmailInterestsUpdateReason": "string",
    "PrimaryAddress": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Street": "string",
        "City": "string",
        "State": "string",
        "PostalCode": "string",
        "Country": "string",
        "IsPrimary": "true",
        "IsBad": "true",
        "StateAbbreviation": "string",
        "CountryCode": "string",
    },
    "EngagementScore": "Low",
    "DonorSearchInfo": {
        "Id": 0,
        "GenerosityScore": "Low",
        "AnnualFundLikelihood": "Low",
        "MajorGiftLikelihood": "Low",
        "PlannedGiftLikelihood": "Low",
        "Quality": "Low",
        "LargestGiftMin": 0,
        "LargestGiftMax": 0,
        "WealthAskMin": 0,
        "WealthAskMax": 0,
        "BusinessExecutive": "true",
        "NamesScreened": "string",
        "DateTimeScreenedUtc": "string",
    },
    "AddressIds": [0],
    "EmailIds": [0],
    "PhoneIds": [0],
    "CustomValues": [
        {"FieldId": 0, "Value": {"Id": 0, "Value": "string"}},
        {"FieldId": 0, "Values": [{"Id": 0, "Value": "string"}]},
    ],
    "AuditTrail": {
        "CreatedDate": "2020-09-08T16:06:59.945Z",
        "CreatedName": "string",
        "LastModifiedDate": "2020-09-08T16:06:59.945Z",
        "LastModifiedName": "string",
    },
}

TEST_GET_CONSTITUENTS = {
    "Total": 0,
    "TotalFiltered": 0,
    "Start": 0,
    "ResultCount": 0,
    "Results": [
        {
            "Id": 0,
            "AccountNumber": 0,
            "IsInHousehold": "true",
            "IsHeadOfHousehold": "true",
            "IsFavorite": "true",
            "FullCustomProfileImageId": 0,
            "FullCustomProfileImageUrl": "string",
            "CroppedCustomProfileImageId": 0,
            "CroppedCustomProfileImageUrl": "string",
            "Type": "Individual",
            "Status": "Active",
            "FirstName": "string",
            "LastName": "string",
            "MiddleName": "string",
            "Prefix": "string",
            "Suffix": "string",
            "FullName": "string",
            "InformalName": "string",
            "FormalName": "string",
            "EnvelopeName": "string",
            "RecognitionName": "string",
            "JobTitle": "string",
            "Employer": "string",
            "Website": "string",
            "FacebookId": "string",
            "TwitterId": "string",
            "LinkedInId": "string",
            "Gender": "Male",
            "Birthdate": "2020-09-05",
            "ProfilePictureType": "None",
            "PrimaryEmail": {
                "Id": 0,
                "AccountId": 0,
                "Type": "Home",
                "Value": "user@example.com",
                "IsPrimary": "true",
                "IsBad": "true",
            },
            "PrimaryPhone": {
                "Id": 0,
                "AccountId": 0,
                "Type": "Home",
                "Extension": "string",
                "Number": "string",
                "IsPrimary": "true",
            },
            "HouseholdId": 0,
            "PreferredCommunicationChannel": "Email",
            "CommunicationRestrictions": ["DoNotCall"],
            "CommunicationRestrictionsUpdateReason": "string",
            "EmailInterestType": "All",
            "CustomEmailInterests": [{"Id": 0, "Name": "string"}],
            "EmailInterestsUpdateReason": "string",
            "EngagementScore": "Low",
            "DonorSearchInfo": {
                "Id": 0,
                "GenerosityScore": "Low",
                "AnnualFundLikelihood": "Low",
                "MajorGiftLikelihood": "Low",
                "PlannedGiftLikelihood": "Low",
                "Quality": "Low",
                "LargestGiftMin": 0,
                "LargestGiftMax": 0,
                "WealthAskMin": 0,
                "WealthAskMax": 0,
                "BusinessExecutive": "true",
                "NamesScreened": "string",
                "DateTimeScreenedUtc": "string",
            },
            "AddressIds": [0],
            "EmailIds": [0],
            "PhoneIds": [0],
            "CustomValues": [
                {"FieldId": 0, "Value": {"Id": 0, "Value": "string"}},
                {"FieldId": 0, "Values": [{"Id": 0, "Value": "string"}]},
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-05T01:40:43.035Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-05T01:40:43.035Z",
                "LastModifiedName": "string",
            },
        }
    ],
}

TEST_CREATE_TRANSACTION = {
    "AccountId": 0,
    "Date": "2020-09-08",
    "Amount": 0,
    "Method": "None",
    "CheckNumber": "string",
    "CheckDate": "2020-09-08",
    "CreditCardType": "Visa",
    "CreditCardLastFourDigits": "string",
    "CreditCardExpMonth": 0,
    "CreditCardExpYear": 0,
    "EftAccountType": "Checking",
    "EftLastFourDigits": "string",
    "EftRoutingNumber": "string",
    "InKindDescription": "string",
    "InKindType": "Goods",
    "InKindMarketValue": 0,
    "WalletItemId": 0,
    "IntegrationUrl": "string",
    "Designations": [
        {
            "Amount": 0,
            "Note": "string",
            "AcknowledgementStatus": "true",
            "AcknowledgementInteractionIds": [0],
            "Type": "Donation",
            "NonDeductibleAmount": 0,
            "FundId": 0,
            "QuickbooksAccountId": 0,
            "CampaignId": 0,
            "AppealId": 0,
            "TributeId": 0,
            "SoftCredits": ["null"],
            "CustomValues": [
                {"FieldId": 0, "Value": "string"},
                {"FieldId": 0, "ValueId": 0},
                {"FieldId": 0, "ValueIds": [0]},
            ],
            "Attachments": ["null", "null"],
        },
        {
            "Amount": 0,
            "Note": "string",
            "AcknowledgementStatus": "true",
            "AcknowledgementInteractionIds": [0],
            "Type": "PledgePayment",
            "PledgeId": 0,
            "CustomValues": [
                {"FieldId": 0, "Value": "string"},
                {"FieldId": 0, "ValueId": 0},
                {"FieldId": 0, "ValueIds": [0]},
            ],
            "Attachments": ["null", "null"],
        },
        {
            "RecurringDonationEndDate": "2020-09-08",
            "Amount": 0,
            "Note": "string",
            "AcknowledgementStatus": "true",
            "AcknowledgementInteractionIds": [0],
            "RecurringDonationFrequency": "Weekly",
            "RecurringDonationDay1": 0,
            "RecurringDonationDay2": 0,
            "RecurringDonationStartDate": "2020-09-08",
            "Type": "RecurringDonation",
            "FundId": 0,
            "QuickbooksAccountId": 0,
            "CampaignId": 0,
            "AppealId": 0,
            "TributeId": 0,
            "SoftCredits": ["null"],
            "CustomValues": [
                {"FieldId": 0, "Value": "string"},
                {"FieldId": 0, "ValueId": 0},
                {"FieldId": 0, "ValueIds": [0]},
            ],
            "Attachments": ["null", "null"],
        },
        {
            "Amount": 0,
            "Note": "string",
            "AcknowledgementStatus": "true",
            "AcknowledgementInteractionIds": [0],
            "Type": "RecurringDonationPayment",
            "RecurringDonationId": 0,
            "FundId": 0,
            "QuickbooksAccountId": 0,
            "CampaignId": 0,
            "AppealId": 0,
            "IsExtraPayment": "true",
            "CustomValues": [
                {"FieldId": 0, "Value": "string"},
                {"FieldId": 0, "ValueId": 0},
                {"FieldId": 0, "ValueIds": [0]},
            ],
            "Attachments": ["null", "null"],
        },
    ],
    "Attachments": [
        {"Guid": "string", "Name": "string", "Extension": "string", "Url": "string"},
        {"Id": 0, "Name": "string", "Extension": "string", "Url": "string"},
    ],
}

TEST_GET_TRANSACTION = {
    "Id": 0,
    "TransactionNumber": 0,
    "NonDeductibleAmount": 0,
    "AccountId": 0,
    "Date": "2020-09-08",
    "Amount": 0,
    "Method": "None",
    "CheckNumber": "string",
    "CheckDate": "2020-09-08",
    "CreditCardType": "Visa",
    "CreditCardLastFourDigits": "string",
    "CreditCardExpMonth": 0,
    "CreditCardExpYear": 0,
    "EftAccountType": "Checking",
    "EftLastFourDigits": "string",
    "EftRoutingNumber": "string",
    "InKindDescription": "string",
    "InKindType": "Goods",
    "InKindMarketValue": 0,
    "IntegrationUrl": "string",
    "Designations": [
        {
            "Id": 0,
            "DesignationNumber": 0,
            "TransactionId": 0,
            "Amount": 0,
            "NonDeductibleAmount": 0,
            "Note": "string",
            "AcknowledgementStatus": "true",
            "AcknowledgementInteractionIds": [0],
            "Fund": {"Id": 0, "Name": "string"},
            "QuickbooksAccount": {"Id": 0, "Name": "string"},
            "Campaign": {"Id": 0, "Name": "string"},
            "Appeal": {"Id": 0, "Name": "string"},
            "Tribute": {"Id": 0, "Name": "string"},
            "TributeType": "InHonorOf",
            "SoftCreditIds": [0],
            "AttachmentIds": [0],
            "CustomValues": [
                {"FieldId": 0, "Value": {"Id": 0, "Value": "string"}},
                {"FieldId": 0, "Values": [{"Id": 0, "Value": "string"}]},
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T14:15:24.293Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T14:15:24.293Z",
                "LastModifiedName": "string",
            },
            "Type": "Donation",
        },
        {
            "Id": 0,
            "DesignationNumber": 0,
            "TransactionId": 0,
            "Amount": 0,
            "NonDeductibleAmount": 0,
            "Note": "string",
            "AcknowledgementStatus": "true",
            "AcknowledgementInteractionIds": [0],
            "Fund": {"Id": 0, "Name": "string"},
            "QuickbooksAccount": {"Id": 0, "Name": "string"},
            "Campaign": {"Id": 0, "Name": "string"},
            "Appeal": {"Id": 0, "Name": "string"},
            "Tribute": {"Id": 0, "Name": "string"},
            "TributeType": "InHonorOf",
            "SoftCreditIds": [0],
            "AttachmentIds": [0],
            "CustomValues": [
                {"FieldId": 0, "Value": {"Id": 0, "Value": "string"}},
                {"FieldId": 0, "Values": [{"Id": 0, "Value": "string"}]},
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T14:15:24.293Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T14:15:24.293Z",
                "LastModifiedName": "string",
            },
            "Type": "Pledge",
            "PledgePaymentIds": [0],
            "PledgeInstallments": [
                {"Id": 0, "PledgeId": 0, "Date": "2020-09-08", "Amount": 0}
            ],
            "PledgeBalance": 0,
            "PledgeStatus": "InGoodStanding",
            "PledgeAmountInArrears": 0,
            "PledgeNextInstallmentDate": "2020-09-08",
        },
        {
            "Id": 0,
            "DesignationNumber": 0,
            "TransactionId": 0,
            "Amount": 0,
            "NonDeductibleAmount": 0,
            "Note": "string",
            "AcknowledgementStatus": "true",
            "AcknowledgementInteractionIds": [0],
            "Fund": {"Id": 0, "Name": "string"},
            "QuickbooksAccount": {"Id": 0, "Name": "string"},
            "Campaign": {"Id": 0, "Name": "string"},
            "Appeal": {"Id": 0, "Name": "string"},
            "Tribute": {"Id": 0, "Name": "string"},
            "TributeType": "InHonorOf",
            "SoftCreditIds": [0],
            "AttachmentIds": [0],
            "CustomValues": [
                {"FieldId": 0, "Value": {"Id": 0, "Value": "string"}},
                {"FieldId": 0, "Values": [{"Id": 0, "Value": "string"}]},
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T14:15:24.293Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T14:15:24.293Z",
                "LastModifiedName": "string",
            },
            "Type": "PledgePayment",
            "PledgeId": 0,
        },
        {
            "Id": 0,
            "DesignationNumber": 0,
            "TransactionId": 0,
            "Amount": 0,
            "NonDeductibleAmount": 0,
            "Note": "string",
            "AcknowledgementStatus": "true",
            "AcknowledgementInteractionIds": [0],
            "Fund": {"Id": 0, "Name": "string"},
            "QuickbooksAccount": {"Id": 0, "Name": "string"},
            "Campaign": {"Id": 0, "Name": "string"},
            "Appeal": {"Id": 0, "Name": "string"},
            "Tribute": {"Id": 0, "Name": "string"},
            "TributeType": "InHonorOf",
            "SoftCreditIds": [0],
            "AttachmentIds": [0],
            "CustomValues": [
                {"FieldId": 0, "Value": {"Id": 0, "Value": "string"}},
                {"FieldId": 0, "Values": [{"Id": 0, "Value": "string"}]},
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T14:15:24.293Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T14:15:24.293Z",
                "LastModifiedName": "string",
            },
            "RecurringDonationEndDate": "2020-09-08",
            "RecurringDonationFrequency": "Weekly",
            "RecurringDonationDay1": 0,
            "RecurringDonationDay2": 0,
            "RecurringDonationStartDate": "2020-09-08",
            "Type": "RecurringDonation",
            "RecurringDonationPaymentIds": [0],
            "RecurringDonationNextInstallmentDate": "2020-09-08",
            "RecurringDonationLastPaymentStatus": "AtRisk",
            "RecurringDonationStatus": "Active",
        },
        {
            "Id": 0,
            "DesignationNumber": 0,
            "TransactionId": 0,
            "Amount": 0,
            "NonDeductibleAmount": 0,
            "Note": "string",
            "AcknowledgementStatus": "true",
            "AcknowledgementInteractionIds": [0],
            "Fund": {"Id": 0, "Name": "string"},
            "QuickbooksAccount": {"Id": 0, "Name": "string"},
            "Campaign": {"Id": 0, "Name": "string"},
            "Appeal": {"Id": 0, "Name": "string"},
            "Tribute": {"Id": 0, "Name": "string"},
            "TributeType": "InHonorOf",
            "SoftCreditIds": [0],
            "AttachmentIds": [0],
            "CustomValues": [
                {"FieldId": 0, "Value": {"Id": 0, "Value": "string"}},
                {"FieldId": 0, "Values": [{"Id": 0, "Value": "string"}]},
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T14:15:24.293Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T14:15:24.293Z",
                "LastModifiedName": "string",
            },
            "Type": "RecurringDonationPayment",
            "RecurringDonationId": 0,
        },
    ],
    "AttachmentIds": [0],
    "IsRefunded": "true",
    "RefundIds": [0],
    "AuditTrail": {
        "CreatedDate": "2020-09-08T14:15:24.293Z",
        "CreatedName": "string",
        "LastModifiedDate": "2020-09-08T14:15:24.293Z",
        "LastModifiedName": "string",
    },
}

TEST_GET_TRANSACTIONS = {
    "Total": 0,
    "TotalFiltered": 0,
    "Start": 0,
    "ResultCount": 0,
    "Results": [
        {
            "Id": 0,
            "TransactionNumber": 0,
            "NonDeductibleAmount": 0,
            "AccountId": 0,
            "Date": "2020-09-08",
            "Amount": 0,
            "Method": "None",
            "CheckNumber": "string",
            "CheckDate": "2020-09-08",
            "CreditCardType": "Visa",
            "CreditCardLastFourDigits": "string",
            "CreditCardExpMonth": 0,
            "CreditCardExpYear": 0,
            "EftAccountType": "Checking",
            "EftLastFourDigits": "string",
            "EftRoutingNumber": "string",
            "InKindDescription": "string",
            "InKindType": "Goods",
            "InKindMarketValue": 0,
            "IntegrationUrl": "string",
            "Designations": ["null", "null", "null", "null", "null"],
            "AttachmentIds": [0],
            "IsRefunded": "true",
            "RefundIds": [0],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T16:11:30.821Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T16:11:30.821Z",
                "LastModifiedName": "string",
            },
        }
    ],
}

TEST_CREATE_INTERACTION = {
    "AccountId": 0,
    "Date": "2020-09-08",
    "Note": "string",
    "Channel": "Email",
    "Purpose": "Acknowledgement",
    "Subject": "string",
    "IsInbound": "true",
    "CustomValues": [
        {"FieldId": 0, "Value": "string"},
        {"FieldId": 0, "ValueId": 0},
        {"FieldId": 0, "ValueIds": [0]},
    ],
    "Attachments": [
        {"Guid": "string", "Name": "string", "Extension": "string", "Url": "string"},
        {"Id": 0, "Name": "string", "Extension": "string", "Url": "string"},
    ],
}

TEST_GET_INTERACTION = {
    "Id": 0,
    "Date": "2020-09-08",
    "Note": "string",
    "Channel": "Email",
    "Purpose": "Acknowledgement",
    "Subject": "string",
    "IsInbound": "true",
    "AccountId": 0,
    "TweetId": "string",
    "IsBcc": "true",
    "EmailAddress": "user@example.com",
    "AttachmentIds": [0],
    "LetterAttachmentIds": [0],
    "SurveyLapsedResponses": ["string"],
    "SurveyEmailInteractionId": 0,
    "SurveyResponseInteractionId": 0,
    "CustomValues": [
        {"FieldId": 0, "Value": {"Id": 0, "Value": "string"}},
        {"FieldId": 0, "Values": [{"Id": 0, "Value": "string"}]},
    ],
    "AuditTrail": {
        "CreatedDate": "2020-09-08T15:27:32.767Z",
        "CreatedName": "string",
        "LastModifiedDate": "2020-09-08T15:27:32.767Z",
        "LastModifiedName": "string",
    },
}

TEST_GET_INTERACTIONS = {
    "Total": 0,
    "TotalFiltered": 0,
    "Start": 0,
    "ResultCount": 0,
    "Results": [
        {
            "Id": 0,
            "Date": "2020-09-08",
            "Note": "string",
            "Channel": "Email",
            "Purpose": "Acknowledgement",
            "Subject": "string",
            "IsInbound": "true",
            "AccountId": 0,
            "TweetId": "string",
            "IsBcc": "true",
            "EmailAddress": "user@example.com",
            "AttachmentIds": [0],
            "LetterAttachmentIds": [0],
            "SurveyLapsedResponses": ["string"],
            "SurveyEmailInteractionId": 0,
            "SurveyResponseInteractionId": 0,
            "CustomValues": [
                {"FieldId": 0, "Value": {"Id": 0, "Value": "string"}},
                {"FieldId": 0, "Values": [{"Id": 0, "Value": "string"}]},
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T16:10:36.389Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T16:10:36.389Z",
                "LastModifiedName": "string",
            },
        }
    ],
}



from parsons.etl import Table
import requests
import logging
from parsons.utilities import check_env

logger = logging.getLogger(__name__)

TURBOVOTE_URI = "https://turbovote-admin-http-api.prod.democracy.works/"


class TurboVote(object):
    """
    Instantiate the TurboVote class

    `Args:`
        username: str
            A valid TurboVote username. Not required if ``TURBOVOTE_USERNAME``
            env variable set.
        password: str
            A valid TurboVote password. Not required if ``TURBOVOTE_PASSWORD``
            env variable set.
        subdomain: str
            Your TurboVote subdomain (i.e. ``https://MYORG.turbovote.org``). Not
            required if ``TURBOVOTE_SUBDOMAIN`` env variable set.
    `Returns:`
        class
    """

    def __init__(self, username=None, password=None, subdomain=None):

        self.username = check_env.check("TURBOVOTE_USERNAME", username)
        self.password = check_env.check("TURBOVOTE_PASSWORD", password)
        self.subdomain = check_env.check("TURBOVOTE_SUBDOMAIN", subdomain)
        self.uri = TURBOVOTE_URI

    def _get_token(self):
        # Retrieve a temporary bearer token to access API

        url = self.uri + "login"
        payload = {"username": self.username, "password": self.password}
        r = requests.post(url, data=payload)
        logger.debug(r.url)
        r.raise_for_status()

        return r.json()["id-token"]

    def get_users(self):
        """
        Get users.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.uri + f"partners/{self.subdomain}.turbovote.org/users"

        headers = {"Authorization": f"Bearer {self._get_token()}"}
        r = requests.get(url, headers=headers)
        logger.debug(r)
        r.raise_for_status()
        tbl = Table.from_csv_string(r.text)
        logger.info(f"{tbl.num_rows} users retrieved.")

        return tbl



from parsons import VAN
from test.test_van.responses_people import get_person_response
from test.utils import assert_matching_tables
import requests_mock
import os
import unittest

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestNGPVAN(unittest.TestCase):
    def setUp(self):
        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters", raise_for_status=False)

    @requests_mock.Mocker()
    def test_create_contact_note(self, m):
        m.post(self.van.connection.uri + "people/1/notes", status_code=204)
        self.van.create_contact_note(1, "a", True)

    @requests_mock.Mocker()
    def test_get_contact_notes(self, m):
        json = get_person_response["notes"]
        m.get(self.van.connection.uri + "people/1/notes", json=json)
        assert_matching_tables(json, self.van.get_contact_notes("1"))



import pytest
import os
import paramiko
from contextlib import contextmanager
from copy import deepcopy
from unittest.mock import MagicMock, patch, call
from parsons import Table, SFTP
from parsons.utilities import files as file_util
from test.utils import mark_live_test, assert_matching_tables
from test.fixtures import (  # noqa: F401
    simple_table,
    simple_csv_path,
    simple_compressed_csv_path,
)

#
# Fixtures and constants
#

REMOTE_DIR, CSV, COMPRESSED_CSV, EMPTY, SUBDIR_A, SUBDIR_B, CSV_A, CSV_B = [
    "parsons_test",
    "test.csv",
    "test.csv.gz",
    "empty",
    "subdir_a",
    "subdir_b",
    "test_a.csv",
    "test_b.csv",
]

CSV_PATH, COMPRESSED_CSV_PATH, EMPTY_PATH, SUBDIR_A_PATH, SUBDIR_B_PATH = [
    f"{REMOTE_DIR}/{content}"
    for content in (CSV, COMPRESSED_CSV, EMPTY, SUBDIR_A, SUBDIR_B)
]

CSV_A_PATH, CSV_B_PATH = [
    f"{d}/{content}" for d, content in ((SUBDIR_A_PATH, CSV_B), (SUBDIR_B_PATH, CSV_A))
]

FILE_PATHS = [CSV_PATH, COMPRESSED_CSV_PATH, CSV_A_PATH, CSV_B_PATH]
DIR_PATHS = [REMOTE_DIR, EMPTY_PATH, SUBDIR_A_PATH, SUBDIR_B_PATH]


def sup(sftp, simple_csv_path, simple_compressed_csv_path):  # noqa: F811
    # The setup function creates remote directories and files needed for live tests
    for remote_dir in DIR_PATHS:
        sftp.make_directory(remote_dir)

    for remote_file in FILE_PATHS:
        fixture = simple_compressed_csv_path if "gz" in remote_file else simple_csv_path
        sftp.put_file(fixture, remote_file)


def cleanup(sftp):
    for f in FILE_PATHS:
        sftp.remove_file(f)
    for remote_dir in reversed(DIR_PATHS):
        sftp.remove_directory(remote_dir)


def generate_live_sftp_connection():
    host = os.environ["SFTP_HOST"]
    username = os.environ["SFTP_USERNAME"]
    password = os.environ["SFTP_PASSWORD"]
    return SFTP(host, username, password)


@pytest.fixture
def live_sftp(simple_csv_path, simple_compressed_csv_path, simple_table):  # noqa: F811
    sftp = generate_live_sftp_connection()
    sup(sftp, simple_csv_path, simple_compressed_csv_path)
    yield sftp
    cleanup(sftp)


# This second live_sftp fixture is used for test_get_files so that files are never downloaded and
# mocks can be inspected.
@pytest.fixture
def live_sftp_with_mocked_get(
    simple_csv_path, simple_compressed_csv_path  # noqa: F811
):
    SFTP_with_mocked_get = deepcopy(SFTP)

    # The names of temp files are long arbitrary strings. This makes them predictable.
    def rv(magic_mock):
        return ["foo", "bar", "baz"][magic_mock.call_count]

    get = MagicMock()
    create_temp_file_for_path = MagicMock()
    create_temp_file_for_path.return_value = rv(create_temp_file_for_path)

    # The following two methods are identical to their cognates in the SFTP class, but they
    # substitue a mock for `conn.get` and `files.create_temp_file_for_path`, respectively.
    @contextmanager
    def create_connection(self):
        transport = paramiko.Transport((self.host, self.port))
        pkey = None
        if self.rsa_private_key_file:
            pkey = paramiko.RSAKey.from_private_key_file(self.rsa_private_key_file)

        transport.connect(username=self.username, password=self.password, pkey=pkey)
        conn = paramiko.SFTPClient.from_transport(transport)
        conn.get = get
        yield conn
        conn.close()
        transport.close()

    def get_file(self, remote_path, local_path=None, connection=None):

        if not local_path:
            local_path = create_temp_file_for_path(remote_path)

        if connection:
            connection.get(remote_path, local_path)
        else:
            with self.create_connection() as connection:
                connection.get(remote_path, local_path)

        return local_path

    SFTP_with_mocked_get.create_connection = create_connection
    SFTP_with_mocked_get.get_file = get_file

    sftp = generate_live_sftp_connection()
    sup(sftp, simple_csv_path, simple_compressed_csv_path)

    yield sftp, get

    cleanup(sftp)


#
# Tests
#


def test_credential_validation():
    with pytest.raises(ValueError):
        SFTP(host=None, username=None, password=None)

    with pytest.raises(ValueError):
        SFTP(host=None, username="sam", password="abc123")


@mark_live_test
def test_list_non_existent_directory(live_sftp):
    with pytest.raises(FileNotFoundError):
        live_sftp.list_directory("abc123")


@mark_live_test
def test_list_directory_with_files(live_sftp):
    result = sorted(live_sftp.list_directory(REMOTE_DIR))
    assert result == [EMPTY, SUBDIR_A, SUBDIR_B, CSV, COMPRESSED_CSV]


@mark_live_test
def test_get_non_existent_file(live_sftp):
    with pytest.raises(FileNotFoundError):
        live_sftp.get_file("abc123")


# Helper function
def assert_file_matches_table(local_path, table):
    downloaded_tbl = Table.from_csv(local_path)
    assert_matching_tables(table, downloaded_tbl)


@mark_live_test
def test_get_file(live_sftp, simple_table):  # noqa F811
    local_path = file_util.create_temp_file()
    live_sftp.get_file(CSV_PATH, local_path=local_path)
    assert_file_matches_table(local_path, simple_table)


@mark_live_test
def test_get_table(live_sftp, simple_table):  # noqa F811
    file_util.create_temp_file()
    tbl = live_sftp.get_table(CSV_PATH)
    assert_matching_tables(tbl, simple_table)


@mark_live_test
def test_get_temp_file(live_sftp, simple_table):  # noqa F811
    local_path = live_sftp.get_file(CSV_PATH)
    assert_file_matches_table(local_path, simple_table)


@mark_live_test
@pytest.mark.parametrize("compression", [None, "gzip"])
def test_table_to_sftp_csv(live_sftp, simple_table, compression):  # noqa F811
    host = os.environ["SFTP_HOST"]
    username = os.environ["SFTP_USERNAME"]
    password = os.environ["SFTP_PASSWORD"]
    remote_path = f"{REMOTE_DIR}/test_to_sftp.csv"
    if compression == "gzip":
        remote_path += ".gz"
    simple_table.to_sftp_csv(
        remote_path, host, username, password, compression=compression
    )

    local_path = live_sftp.get_file(remote_path)
    assert_file_matches_table(local_path, simple_table)

    # Cleanup
    live_sftp.remove_file(remote_path)


#
# Helper Functions
#


def assert_results_match_expected(expected, results):
    assert len(results) == len(expected)
    for e in expected:
        assert any([e in r for r in results])


def assert_has_call(mock, args):
    return call(*args) in mock.mock_calls


def assert_has_calls(mock, calls):
    return all([assert_has_call(mock, c) for c in calls])


@mark_live_test
def test_list_files(live_sftp):
    result = sorted(live_sftp.list_files(REMOTE_DIR))
    assert result == [CSV_PATH, COMPRESSED_CSV_PATH]


@mark_live_test
def test_list_files_with_pattern(live_sftp):
    result = live_sftp.list_files(REMOTE_DIR, pattern="gz")
    assert result == [COMPRESSED_CSV_PATH]


@mark_live_test
def test_list_subdirectories(live_sftp):
    result = sorted(live_sftp.list_subdirectories(REMOTE_DIR))
    assert result == [EMPTY_PATH, SUBDIR_A_PATH, SUBDIR_B_PATH]


@mark_live_test
def test_list_subdirectories_with_pattern(live_sftp):
    result = sorted(live_sftp.list_subdirectories(REMOTE_DIR, pattern="sub"))
    assert result == [SUBDIR_A_PATH, SUBDIR_B_PATH]


local_paths = ["foo", "bar"]

# The following are values for the arguments to pass to `get_files` and `walk_tree` as well as the
# strings expected to be found in the returned results.
args_and_expected = {
    "get_files": [
        ({"remote": REMOTE_DIR}, [CSV_PATH, COMPRESSED_CSV_PATH]),
        ({"remote": [SUBDIR_A_PATH, SUBDIR_B_PATH]}, [CSV_B_PATH, CSV_A_PATH]),
        (
            {"remote": SUBDIR_B_PATH, "files_to_download": [CSV_B_PATH]},
            [CSV_A_PATH, CSV_B_PATH],
        ),
        ({"remote": [SUBDIR_A_PATH, SUBDIR_B_PATH], "pattern": "a"}, [CSV_B_PATH]),
    ],
    "walk_tree": [
        (
            [REMOTE_DIR],
            {"download": False, "dir_pattern": SUBDIR_A},
            [[SUBDIR_A], [COMPRESSED_CSV, CSV, CSV_B]],
        ),
        (
            [REMOTE_DIR],
            {"download": False, "file_pattern": CSV_B},
            [[SUBDIR_A, SUBDIR_B, EMPTY], [CSV_B]],
        ),
        (
            [REMOTE_DIR],
            {"download": False, "dir_pattern": SUBDIR_A, "file_pattern": CSV_B},
            [[SUBDIR_A], [CSV_B]],
        ),
        (
            [REMOTE_DIR],
            {"download": False, "max_depth": 1},
            [[EMPTY, SUBDIR_A, SUBDIR_B], [CSV, COMPRESSED_CSV]],
        ),
    ],
}


@mark_live_test
def test_get_files_calls_get_to_write_to_provided_local_paths(
    live_sftp_with_mocked_get,
):
    live_sftp, get = live_sftp_with_mocked_get
    results = live_sftp.get_files(
        remote=[SUBDIR_A_PATH, SUBDIR_B_PATH], local_paths=local_paths
    )
    assert get.call_count == 2
    calls = [call(CSV_A_PATH, local_paths[0]), call(CSV_B_PATH, local_paths[1])]
    assert_has_calls(get, calls)
    assert_results_match_expected(local_paths, results)


@mark_live_test
@pytest.mark.parametrize("kwargs,expected", args_and_expected["get_files"])
def test_get_files_calls_get_to_write_temp_files(
    kwargs, expected, live_sftp_with_mocked_get
):
    live_sftp, get = live_sftp_with_mocked_get
    live_sftp.get_files(**kwargs)
    assert get.call_count == len(expected)
    calls = [call(e, local_paths[i]) for i, e in enumerate(expected)]
    assert_has_calls(get, calls)


@mark_live_test
def test_get_files_raises_error_when_no_file_source_is_provided(live_sftp):
    with pytest.raises(ValueError):
        live_sftp.get_files()


@mark_live_test
@patch("parsons.sftp.SFTP.get_file")
def test_get_files_with_files_paths_mismatch(get_file, live_sftp):
    live_sftp.get_files(files_to_download=[CSV_A_PATH], local_paths=local_paths)
    assert get_file.call_args[1]["local_path"] is None


@mark_live_test
@pytest.mark.parametrize("args,kwargs,expected", args_and_expected["walk_tree"])
def test_walk_tree(args, kwargs, expected, live_sftp_with_mocked_get):
    live_sftp, get = live_sftp_with_mocked_get
    results = live_sftp.walk_tree(*args, **kwargs)
    # `results` will be a list of first dirs then files, as will `expected`
    for res, expect in zip(results, expected):
        assert_results_match_expected(expect, res)


# Stuff that is tested by the live_sftp fixture, so no need to test explicitly:
# test_make_directory
# test_put_file
# test_put_file_compressed
# def test_remove_file
# def test_remove_directory



"""Implements client routine to allow execution of TargetSmart SmartMatch
workflows.

TargetSmart SmartMatch API doc:
https://docs.targetsmart.com/developers/tsapis/v2/service/smartmatch.html
"""

import gzip
import logging
import shutil
import tempfile
import time
import uuid

import petl
import requests

from parsons import Table

logger = logging.getLogger(__name__)

VALID_FIELDS = [
    "voterbase_id",
    "smartvan_id",
    "voter_id",
    "exact_track",
    "full_name",
    "first_name_combined",
    "first_name",
    "middle_name",
    "last_name",
    "name_suffix",
    "address1",
    "address2",
    "city",
    "state",
    "zip",
    "age",
    "gender",
    "dob",
    "phone",
    "email",
    "latitude",
    "longitude",
]

INTERNAL_JOIN_ID = "matchback_id"
INTERNAL_JOIN_ID_CONFLICT = "__matchback_id"


class SmartMatchError(Exception):
    """Raised when SmartMatch workflow processing fails."""


def _smartmatch_upload(url, fname):
    logger.info(f"Uploading {fname} to {url} to begin SmartMatch workflow execution.")
    with open(fname, "rb") as reader:
        response_2 = requests.put(url, data=reader, headers={"Content-Type": ""})

    response_2.raise_for_status()


def _smartmatch_download(url, writer):
    with requests.get(url, stream=True) as response:
        response.raise_for_status()
        for chunk in response.iter_content(chunk_size=8192):
            writer.write(chunk)


def _add_join_id(input_table):
    """`matchback_id` is added to the raw input table so the results can later be
    joined back. Integer sequence values are used. If the column already exists
    in the raw input, it is renamed to `__matchback_id` and restored after
    result join.
    """
    if INTERNAL_JOIN_ID in input_table.fieldnames():
        input_table = input_table.rename(INTERNAL_JOIN_ID, INTERNAL_JOIN_ID_CONFLICT)

    return input_table.addrownumbers(field=INTERNAL_JOIN_ID)


def _prepare_input(intable, tmpdir):
    valid = VALID_FIELDS + [INTERNAL_JOIN_ID]
    supported = set(intable.fieldnames()) & set(valid)
    if not supported:
        raise SmartMatchError(
            "No supported field identifiers were found in the input table."
            f" Expecting one or more from: {VALID_FIELDS}"
        )
    return intable.cut(*supported)


class SmartMatch:
    """
    Works as a mixin to the TargetSmartAPI class.
    """

    def __init__(self):
        # Set by TargetSmartAPI constructor
        self.connection = None

    def _smartmatch_poll(self, poll_url, submit_filename):
        download_url = None
        while True:
            poll_response = requests.get(
                poll_url,
                {"filename": submit_filename},
                headers=self.connection.headers,
            )

            if poll_response.ok:
                poll_info = poll_response.json()

                if poll_info["error"]:
                    raise SmartMatchError(poll_info["error"])

                download_url = poll_info["url"]
                if download_url:
                    break
            time.sleep(60 * 2.5)
        return download_url

    def smartmatch(
        self,
        input_table,
        max_matches=1,
        include_email=False,
        include_landline=False,
        include_wireless=False,
        include_voip=False,
        tmp_location=None,
        keep_smartmatch_input_file=False,
        keep_smartmatch_output_gz_file=False,
    ):
        """Submit the contact list records available in the Parsons table ``input_table`` to
        TargetSmart SmartMatch.

        * `SmartMatch overview <https://docs.targetsmart.com/my_tsmart/smartmatch/overview.html>`_
        * `SmartMatch API doc <https://docs.targetsmart.com/developers/tsapis/v2/service/smartmatch.html>`_
        * `Supported input header field identifiers <https://docs.targetsmart.com/developers/tsapis/v2/service/smartmatch.html#supported-field-identifiers>`_

        Your application provides a contact list which will be matched to
        TargetSmart’s database of voting age individuals.

        `TargetSmart Client Services <mailto:support@targetsmart.com>`_
        provisions SmartMatch for your API key, configuring the fields from the
        TargetSmart Data Dictionary that will be appended to each matched
        record.

        This method blocks until TargetSmart has completed the remote workflow
        execution. The execution time can take minutes to hours to complete
        depending on the file size, the types of field identifiers present, and
        TargetSmart system load. SmartMatch executions cannot be canceled once
        submitted to TargetSmart.

        Since Parsons Petl tables are lazy, the SmartMatch output file is always
        retained in ``tmp_location``. If your Parsons-based ETL workflow fails
        downstream it may be beneficial to recover the raw SmartMatch output
        from this location. You may delete this data when it is no longer
        needed.

        `Args:`
            input_table: Parsons or Petl table
                A Parsons table with `header field names supported by SmartMatch <https://docs.targetsmart.com/developers/tsapis/v2/service/smartmatch.html#supported-field-identifiers>`_. Required.
            max_matches: int
                By default only a single best match is returned for an input record. Increase to return additional potentially accurate matches for each input record. Value between 1-10. Default of 1.
            include_email: bool
                Set to True to include appended email values for matched records. This is only applicable if your TargetSmart account is configued to return email data. Additional charges may apply if True. Default of False.
            include_landline: bool
                Set to True to include appended landline phone number values for matched records. This is only applicable if your TargetSmart account is configued to return landline phone data. Additional charges may apply if True. Default of False.
            include_wireless: bool
                Set to True to include appended wireless phone number values for matched records. This is only applicable if your TargetSmart account is configued to return wireless phone data. Additional charges may apply if True. Default of False.
            include_voip: bool
                Set to True to include appended VOIP phone number values for matched records. This is only applicable if your TargetSmart account is configued to return VOIP phone data. Additional charges may apply if True. Default of False.
            tmp_location: str
                Optionally provide a local directory path where input/output CSV files will be stored. Useful to recover CSV output if downstream ETL processing fails. If not specified, a system tmp location is used. Default of None.
            keep_smartmatch_input_file: bool
                Optionally keep the CSV input file that is uploaded in ``tmp_location`` for later use. Default of False.
            keep_smartmatch_output_gz_file: bool
                Optionally keep the gzip compressed output file in ``tmp_location`` for later use. The uncompressed output file is always retained in ``tmp_location``. Default of False
        `Returns:`
            Parsons Table
                A Parsons table wrapping the SmartMatch execution output file records. Each record will
                include the input record fields followed by columns named ``tsmart_match_code``, a
                match indicator, ``vb.voterbase_id``, and zero or more additional data
                element fields based on your TargetSmart account configuration.
                See :ref:`parsons-table` for output options.
        """  # noqa

        # If `input_table` is a Parsons table, convert it to a Petl table.
        if hasattr(input_table, "table"):
            input_table = input_table.table

        url = self.connection.uri + "service/smartmatch"
        poll_url = f"{url}/poll"

        if not input_table:
            raise ValueError(
                "Missing `input_table`. A Petl table must be provided with"
                " valid input rows."
            )

        if not hasattr(input_table, "tocsv"):
            raise ValueError("`input_table` isn't a valid table.")

        if int(max_matches) > 10:
            raise ValueError("max_matches cannot be greater than 10")

        if not tmp_location:
            tmp_location = tempfile.mkdtemp()

        logger.info("Preparing data for SmartMatch submission.")
        input_table = _add_join_id(input_table)
        dataprep_table = _prepare_input(input_table, tmp_location)
        # Unique execution label for each submission
        submit_filename = f"tmc_{str(uuid.uuid4())[0:10]}.csv"

        # An initial api.targetsmart.com request is performed to register the
        # job execution. The response returns a presigned S3 url where data
        # records will be uploaded.
        response_1 = requests.get(
            url,
            {
                "filename": submit_filename,
                "include_email": include_email,
                "include_landline": include_landline,
                "include_wireless": include_wireless,
                "include_voip": include_voip,
                "max_matches": max_matches,
                "format": "gzip",
            },
            headers=self.connection.headers,
        )
        response_1.raise_for_status()
        response_1_info = response_1.json()
        if response_1_info["error"]:
            raise SmartMatchError(
                "SmartMatch workflow registration failed. Error:"
                f" {response_1_info['error']}"
            )

        logger.info(
            "The SmartMatch workflow registration was successful for file name"
            f" {submit_filename}."
        )

        # Write Petl table to CSV and upload for SmartMatch to process
        with tempfile.NamedTemporaryFile(
            mode="w+",
            encoding="utf8",
            newline="\n",
            prefix="smartmatch_input",
            suffix=".csv",
            dir=tmp_location,
            delete=not keep_smartmatch_input_file,
        ) as tmp:
            dataprep_table.tocsv(tmp.name, encoding="utf8")
            tmp.flush()
            _smartmatch_upload(response_1_info["url"], tmp.name)

        logger.info(
            "The SmartMatch workflow execution has been submitted using file"
            f" name '{submit_filename}'. Now polling for results which can take"
            " minutes/hours depending on data size and queuing."
        )

        # Poll SmartMatch endpoint waiting for workflow completion
        download_url = self._smartmatch_poll(poll_url, submit_filename)

        # Download SmartMatch .csv.gz results, decompress, and Petl table wrap.
        # The final tmp file cannot be deleted due to Petl tables being lazy.
        with tempfile.NamedTemporaryFile(
            prefix="smartmatch_output",
            suffix=".csv.gz",
            dir=tmp_location,
            delete=not keep_smartmatch_output_gz_file,
        ) as tmp_gz:
            with tempfile.NamedTemporaryFile(
                prefix="smartmatch_output",
                suffix=".csv",
                dir=tmp_location,
                delete=False,
            ) as tmp_csv:
                logger.info(
                    f"Downloading the '{submit_filename}' SmartMatch results to"
                    f" {tmp_gz.name}."
                )
                _smartmatch_download(download_url, tmp_gz)
                tmp_gz.flush()

                logger.info("Decompressing results")
                with gzip.open(tmp_gz.name, "rb") as gz_reader:
                    shutil.copyfileobj(gz_reader, tmp_csv)
                tmp_csv.flush()

                raw_outtable = petl.fromcsv(  # pylint: disable=no-member
                    tmp_csv.name, encoding="utf8"
                ).convert(INTERNAL_JOIN_ID, int)
                logger.info(
                    "SmartMatch remote execution successful. Joining results to"
                    " input table."
                )
                outtable = (
                    petl.leftjoin(  # pylint: disable=no-member
                        input_table,
                        raw_outtable,
                        key=INTERNAL_JOIN_ID,
                        tempdir=tmp_location,
                    )
                    .sort(key=INTERNAL_JOIN_ID)
                    .cutout(INTERNAL_JOIN_ID)
                )
                if INTERNAL_JOIN_ID_CONFLICT in input_table.fieldnames():
                    input_table = input_table.rename(
                        INTERNAL_JOIN_ID_CONFLICT, INTERNAL_JOIN_ID
                    )

                return Table(outtable)
